/* chunk_8svx_body.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_8svx_body.hh"
#include "chunk_vhdr.hh"

int Chunk8SVXBODY::parse_chunk( const std::vector< char > &data )
{
    //auto chunklen = data.size();

    if ( m_chunk_type != CHUNKNAME( "BODY" ) ) {
        return -EINVAL;
    }

    return 0;
}

json_t *Chunk8SVXBODY::dump_as_json( const shared_properties_t &shared_props,
                                     bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "BODY" ) );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

int Chunk8SVXBODY::decompressFDE( std::vector< int8_t > &result,
                                  const std::vector< int8_t > &input ) const
{
    int8_t fibonacciDeltas[] = { -34, -21, -13, -8, -5, -3, -2, -1,
                                 0, 1, 2, 3, 5, 8, 13, 21 };
    size_t len = input.size();

    if ( len < 2 ) {
        LOG_CRITICAL( "invalid size of FDE data" );
        return -EINVAL;
    }

    int8_t current_sample = input[1];

    len -= 2;
    len *= 2;

    result.resize( len );

    for ( size_t i = 0; i < len; i++ ) {
        if ( 2 + ( i / 2 ) >= input.size() ) {
            LOG_CRITICAL( "insufficient input data" );
            return -EINVAL;
        }

        uint8_t encoded_delta = (uint8_t)input[ 2 + ( i / 2 ) ];

        if ( i & 1 ) {
            encoded_delta &= 0xf;
        } else {
            encoded_delta >>= 4;
        }

        current_sample += fibonacciDeltas[ encoded_delta ];
        result[i] = current_sample;
    }

    return 0;
}

int Chunk8SVXBODY::decode( const std::shared_ptr< ChunkVHDR > &vhdr,
                           std::vector< std::vector< int8_t > > &all_samples ) const
{
    const uint8_t octaves = vhdr->no_of_octaves();
    uint32_t no_of_one_shot_samples = vhdr->one_shot_samples();
    uint32_t no_of_repeat_samples = vhdr->repeat_samples();
    const uint8_t compression = vhdr->compression();

    size_t current_offset = 0;
    std::vector< int8_t > input;

    for ( auto &ch : m_raw_data ) {
        input.push_back( (int8_t)ch );
    }

    if ( compression == ChunkVHDR::COMPRESSION_FDE ) {
        std::vector< int8_t > decoded;
        int res = decompressFDE( decoded, input );
        if ( res != 0 ) {
            return res;
        }
        input = decoded;
    } else if ( compression != ChunkVHDR::COMPRESSION_NONE ) {
        LOG_CRITICAL( "compression %u not supported", compression );
        return -ENOTSUP;
    }

    for ( uint8_t octave = 0; octave < octaves; octave++ ) {
        std::vector< int8_t > one_shot_samples;
        std::vector< int8_t > repeat_samples;

        for ( uint32_t i = 0; i < no_of_one_shot_samples; i++ ) {
            if ( current_offset >= input.size() ) {
                LOG_CRITICAL( "samples missing in body" );
                break;
            }

            int8_t v = input.at( current_offset++ );
            one_shot_samples.push_back( v );
        }

        for ( uint32_t i = 0; i < no_of_repeat_samples; i++ ) {
            if ( current_offset >= input.size() ) {
                LOG_CRITICAL( "samples missing in body" );
                break;
            }

            int8_t v = input.at( current_offset++ );
            repeat_samples.push_back( v );
        }

        all_samples.push_back( one_shot_samples );
        all_samples.push_back( repeat_samples );

        no_of_one_shot_samples *= 2;
        no_of_repeat_samples *= 2;
    }

    if ( current_offset < input.size() ) {
        LOG_VERBOSE( "still some bytes left in body" );
    }

    return 0;
}
