/* chunk_ammd_form.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_ammd_form.hh"
#include "chunk_form.hh"
#include <algorithm>

int ChunkAMMDFORM::read_into_raw( const char *buf,
                                 size_t buflen,
                                 size_t &raw_chunklen )
{
    if ( buflen < 8 ) {
        return -EINVAL;
    }

    m_chunk_type = betoh_u32( get_u32( buf, 0 ) );

    auto chunklen = betoh_u32( get_u32( buf, 4 ) );

    // AMMD forms use an invalid form length which includes the length
    // field as well

    if ( chunklen + 4 > buflen ||
         chunklen < 4 ) {
        LOG_DEBUG( "invalid chunklen %u/%u", chunklen, buflen );
        return -EINVAL;
    }

    m_raw_data.resize( chunklen - 4 );
    std::copy_n( buf + 8, chunklen - 4,
                 m_raw_data.begin() );

    if ( chunklen & 1 ) {
        raw_chunklen = 4 + chunklen + 1;
    } else {
        raw_chunklen = 4 + chunklen;
    }

    LOG_DEBUG( "Read chunk '%.*s' of %u bytes into raw_data for parsing", 4, buf, chunklen );

    return 0;
}

int ChunkAMMDFORM::parse_chunk( const std::vector< char > &data )
{
    auto buflen = data.size();
    const char *buf = &data[0];

    if ( buflen < 4 ) {
        LOG_DEBUG( "too short for AMMD FORM" );
        return -EINVAL;
    }

    if ( m_chunk_type != CHUNKNAME( "FORM" ) ) {
        LOG_DEBUG( "not a AMMD FORM chunk" );
        return -EINVAL;
    }

    LOG_DEBUG( "Found AMMD FORM chunk" );
    LOG_DEBUG( "parse %u bytes...", buflen );

    m_ammdform_type = betoh_u32( get_u32( buf, 0 ) );
    auto ammdform_string = std::string( buf, 4 );

    LOG_DEBUG( "Found AMMD FORM type %s", ammdform_string.c_str() );
    
    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        //auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        LOG_DEBUG( "Next chunk %.*s", 4, buf );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length: %u/%u", chunklen, buflen );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "BODY",
                                                                                    "WFMS",
                                                                                    "MDLS",
                                                                                    "INMS",
                                                                                    "SDTA" } );
        if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
            LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
            if ( fail_on_unsupported() ) {
                return -EINVAL;
            }
        }

        LOG_DEBUG( "got generic chunk %.*s of %u bytes", 4, buf, chunk_parsed_len );

        m_chunks.push_back( chunk );

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    LOG_DEBUG( "Done parsing AMMD FORM chunk" );

    return 0;
}

std::string ChunkAMMDFORM::dump_as_readable_text( const shared_properties_t &shared_props )
{
    std::string res;

    for ( auto &chunk : m_chunks ) {
        res += chunk->dump_as_readable_text( shared_props );
    }

    return res;
}

json_t *ChunkAMMDFORM::dump_as_json( const shared_properties_t &shared_props,
                                     bool include_raw ) const
{
    LOG_DEBUG("called");

    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "FORM" ) );

    json_object_set_new( jroot, "subtype", json_string( CHUNKTYPE2NAME( m_ammdform_type ).c_str() ) );

    auto jarray = json_array();

    for ( auto &chunk : m_chunks ) {
        auto jentry = chunk->dump_as_json( shared_props, include_raw );

        json_array_append_new( jarray, jentry );
    }

    json_object_set_new( jroot, "content", jarray );

    return jroot;
}

int ChunkAMMDFORM::dump_as_binary( const shared_properties_t &shared_props,
                               std::list< IFFBinaryResult > &output,
                               bool include_raw ) const
{
    for ( auto &chunk : m_chunks ) {
        if ( chunk->dump_as_binary( shared_props,
                                    output,
                                    include_raw ) != 0 ) {
            LOG_DEBUG( "failed to dump binary data" );
        }
    }

    return 0;
}
