/* chunk_anhd.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_anhd.hh"

int ChunkANHD::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();

    if ( m_chunk_type != CHUNKNAME( "ANHD" ) ) {
        return -EINVAL;
    }

    const char *buf = &data[0];

    if ( chunklen != 40 ) {
        LOG_DEBUG( "invalid ANHD chunk size: %u", chunklen );
        return -EINVAL;
    }

    m_operation = get_u8( buf, 0 );
    m_mask = get_u8( buf, 1 );

    m_w = betoh_u16( get_u16( buf, 2 ) );
    m_h = betoh_u16( get_u16( buf, 4 ) );
    m_x = (int16_t)betoh_u16( get_u16( buf, 6 ) );
    m_y = (int16_t)betoh_u16( get_u16( buf, 8 ) );

    m_abstime = betoh_u32( get_u32( buf, 10 ) );
    m_reltime = betoh_u32( get_u32( buf, 14 ) );

    m_interleave = get_u8( buf, 18 );

    m_bits = betoh_u32( get_u32( buf, 20 ) );

    return 0;
}

json_t *ChunkANHD::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( CHUNKTYPE2NAME( m_chunk_type ).c_str() ) );

    json_object_set_new( jroot, "operationValue", json_integer( m_operation ) );
    switch ( m_operation ) {
        case OPERATION_DIRECT:
            json_object_set_new( jroot, "operation", json_string( "direct" ) );
            break;
        case OPERATION_XOR:
            json_object_set_new( jroot, "operation", json_string( "xor" ) );
            break;
        case OPERATION_LONG_DELTA:
            json_object_set_new( jroot, "operation", json_string( "longDelta" ) );
            break;
        case OPERATION_SHORT_DELTA:
            json_object_set_new( jroot, "operation", json_string( "shortDelta" ) );
            break;
        case OPERATION_SHORT_LONG_DELTA:
            json_object_set_new( jroot, "operation", json_string( "shortLongDelta" ) );
            break;
        case OPERATION_VERTICAL_DELTA:
            json_object_set_new( jroot, "operation", json_string( "verticalDelta" ) );
            break;
        case OPERATION_SHORT_LONG_VDELTA:
            json_object_set_new( jroot, "operation", json_string( "shortLongVerticalDelta" ) );
            break;
        case OPERATION_DELTA8:
            json_object_set_new( jroot, "operation", json_string( "delta8" ) );
            break;
        default:
            json_object_set_new( jroot, "operation", json_string( "invalid" ) );
            break;
    }

    json_object_set_new( jroot, "mask", json_integer( m_mask ) );

    json_object_set_new( jroot, "width", json_integer( m_w ) );
    json_object_set_new( jroot, "height", json_integer( m_h ) );
    json_object_set_new( jroot, "x", json_integer( m_x ) );
    json_object_set_new( jroot, "y", json_integer( m_y ) );

    json_object_set_new( jroot, "absTime", json_integer( m_abstime ) );
    json_object_set_new( jroot, "relTime", json_integer( m_reltime ) );

    json_object_set_new( jroot, "interleave", json_integer( m_interleave ) );
    json_object_set_new( jroot, "bits", json_integer( m_bits ) );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

uint8_t ChunkANHD::operation() const
{
    return m_operation;
}

uint8_t ChunkANHD::mask() const
{
    return m_mask;
}

uint16_t ChunkANHD::w() const
{
    return m_w;
}

uint16_t ChunkANHD::h() const
{
    return m_h;
}

int16_t ChunkANHD::x() const
{
    return m_x;
}

int16_t ChunkANHD::y() const
{
    return m_y;
}

uint32_t ChunkANHD::abstime() const
{
    return m_abstime;
}

uint32_t ChunkANHD::reltime() const
{
    return m_reltime;
}

uint8_t ChunkANHD::interleave() const
{
    return m_interleave;
}

uint32_t ChunkANHD::bits() const
{
    return m_bits;
}
