/* chunk_bmhd.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_bmhd.hh"

int ChunkBMHD::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();

    if ( m_chunk_type != CHUNKNAME( "BMHD" ) ) {
        return -EINVAL;
    }

    const char *buf = &data[0];

    if ( chunklen != 20 ) {
        LOG_DEBUG( "invalid BMHD chunk size: %u", chunklen );
        return -EINVAL;
    }

    m_width = betoh_u16( get_u16( buf, 0 ) );
    m_height = betoh_u16( get_u16( buf, 2 ) );
    m_pos_x = (int16_t)betoh_u16( get_u16( buf, 4 ) );
    m_pos_y = (int16_t)betoh_u16( get_u16( buf, 6 ) );
    m_no_of_planes = get_u8( buf, 8 );
    m_masking = get_u8( buf, 9 );
    m_compression = get_u8( buf, 10 );
    m_cmap_flags = get_u8( buf, 11 );
    m_transparency_color = betoh_u16( get_u16( buf, 12 ) );
    m_xaspect = get_u8( buf, 14 );
    m_yaspect = get_u8( buf, 15 );
    m_page_width = betoh_u16( get_u16( buf, 16 ) );
    m_page_height = betoh_u16( get_u16( buf, 18 ) );

    return 0;
}

json_t *ChunkBMHD::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "BMHD" ) );
    json_object_set_new( jroot, "width", json_integer( m_width ) );
    json_object_set_new( jroot, "height", json_integer( m_height ) );
    json_object_set_new( jroot, "x", json_integer( m_pos_x ) );
    json_object_set_new( jroot, "y", json_integer( m_pos_y ) );
    json_object_set_new( jroot, "nPlanes", json_integer( m_no_of_planes ) );
    json_object_set_new( jroot, "transparencyColor", json_integer( m_transparency_color ) );
    json_object_set_new( jroot, "xAspect", json_integer( m_xaspect ) );
    json_object_set_new( jroot, "yAspect", json_integer( m_yaspect ) );
    json_object_set_new( jroot, "pageWidth", json_integer( m_page_width ) );
    json_object_set_new( jroot, "pageHeight", json_integer( m_page_height ) );

    json_object_set_new( jroot, "maskValue", json_integer( m_masking ) );
    switch ( m_masking ) {
        case MASK_NONE:
            json_object_set_new( jroot, "mask", json_string( "none" ) );
            break;
        case MASK_HAS_MASK:
            json_object_set_new( jroot, "mask", json_string( "hasMask" ) );
            break;
        case MASK_HAS_TRANSPARENCY_COLOR:
            json_object_set_new( jroot, "mask", json_string( "hasTransparencyColor" ) );
            break;
        case MASK_LASSO:
            json_object_set_new( jroot, "mask", json_string( "lasso" ) );
            break;
        default:
            json_object_set_new( jroot, "mask", json_string( "invalid" ) );
            break;
    }

    json_object_set_new( jroot, "compressionValue", json_integer( m_compression ) );
    switch ( m_compression ) {
        case COMPRESSION_NONE:
            json_object_set_new( jroot, "compression", json_string( "none" ) );
            break;
        case COMPRESSION_BYTERUN1:
            json_object_set_new( jroot, "compression", json_string( "byterun1" ) );
            break;
        default:
            json_object_set_new( jroot, "compression", json_string( "invalid" ) );
            break;
    }

    json_object_set_new( jroot, "cmapFlagsValue", json_integer( m_cmap_flags ) );

    auto cmapFlags = json_array();
    if ( m_cmap_flags & CMAP_FLAG_8BIT ) {
        json_array_append_new( cmapFlags, json_string( "eightBitColors" ) );
    }
    json_object_set_new( jroot, "cmapFlags", cmapFlags );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

uint16_t ChunkBMHD::width() const
{
    return m_width;
}

uint16_t ChunkBMHD::height() const
{
    return m_height;
}

uint8_t ChunkBMHD::nplanes() const
{
    return m_no_of_planes;
}

uint8_t ChunkBMHD::masking() const
{
    return m_masking;
}

uint8_t ChunkBMHD::compression() const
{
    return m_compression;
}

uint8_t ChunkBMHD::cmap_flags() const
{
    return m_cmap_flags;
}

uint16_t ChunkBMHD::transparencyColor() const
{
    return m_transparency_color;
}

size_t ChunkBMHD::line_pitch() const
{
    size_t pixels_per_line =  ( ( m_width + 15 ) / 16 ) * 16;

    return pixels_per_line;
}
