/* chunk_body.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_body.hh"
#include "chunk_bmhd.hh"

int ChunkBODY::parse_chunk( const std::vector< char > &data )
{
    //auto chunklen = data.size();

    if ( m_chunk_type != CHUNKNAME( "BODY" ) ) {
        return -EINVAL;
    }

    return 0;
}

json_t *ChunkBODY::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "BODY" ) );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

int ChunkBODY::decode( const std::shared_ptr< ChunkBMHD > &bmhd,
                       std::vector< uint32_t > &res ) const
{
    std::vector< std::vector< uint8_t > > planes;
    std::vector< uint8_t > mask;

    int decode_res = get_planes( bmhd, planes, mask );
    if ( decode_res != 0 ) {
        return decode_res;
    }

    decode_res = planes_to_pixels( bmhd, planes, mask, res );
    if ( decode_res != 0 ) {
        return decode_res;
    }

    return 0;
}

size_t ChunkBODY::decode_row( size_t start_offset, size_t output_bytes,
                              uint8_t compression, std::vector< uint8_t > &res ) const
{
    size_t offset = start_offset;

    if ( compression == ChunkBMHD::COMPRESSION_NONE ) {
        while ( output_bytes > 0 ) {
            if ( offset >= m_raw_data.size() ) {
                LOG_CRITICAL( "body too small" );
                return -EINVAL;
            }
            uint8_t v = m_raw_data.at( offset );
            offset++;

            res.push_back( v );
            output_bytes--;
        }
    } else if ( compression == ChunkBMHD::COMPRESSION_BYTERUN1 ) {
        while ( output_bytes > 0 ) {
            if ( offset >= m_raw_data.size() ) {
                LOG_CRITICAL( "body too small" );
                return -EINVAL;
            }
            uint8_t v = m_raw_data.at( offset );
            offset++;

            if ( v == 0x80 ) {
            } else if ( v < 0x80 ) {
                //LOG_DEBUG( "copy %u bytes", v + 1 );

                for ( uint8_t c = 0; c < v + 1; c++ ) {
                    uint8_t v2 = m_raw_data.at( offset );
                    offset++;

                    if ( output_bytes == 0 ) {
                        LOG_CRITICAL( "too many bytes written" );
                        return -EINVAL;
                    }

                    res.push_back( v2 );
                    output_bytes--;
                }
            } else {
                uint8_t v2 = m_raw_data.at( offset );
                offset++;

                //LOG_DEBUG( "repeat byte %u times", (uint8_t)((~v) + 2) );

                for ( uint8_t c = (~v) + 2;
                      c > 0;
                      c-- ) {
                    if ( output_bytes == 0 ) {
                        LOG_CRITICAL( "too many bytes written" );
                        return -EINVAL;
                    }

                    res.push_back( v2 );
                    output_bytes--;
                }
            }
        }
    } else {
        LOG_CRITICAL( "unsupported compression" );
        return -ENOTSUP;
    }

    return offset - start_offset;
}

int ChunkBODY::get_planes( const std::shared_ptr< ChunkBMHD > &bmhd,
                           std::vector< std::vector< uint8_t > > &planes,
                           std::vector< uint8_t > &mask ) const
{
    const uint8_t nplanes = bmhd->nplanes();

    size_t current_offset = 0;
    size_t row_bytes = ( ( bmhd->width() + 15 ) / 16 ) * 2;

    bool failed = false;

    planes.resize( nplanes );

    for ( uint16_t y = 0; y < bmhd->height() && ! failed; y++ ) {
        LOG_DEBUG( "decode row %u", y );

        for ( uint8_t p = 0; p < nplanes && ! failed; p++ ) {
            size_t res = decode_row( current_offset,
                                     row_bytes,
                                     bmhd->compression(),
                                     planes[p] );
            if ( res == 0 ) {
                LOG_CRITICAL( "invalid body" );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                } else {
                    failed = true;
                    break;
                }
            }

            current_offset += res;

            LOG_DEBUG( "plane %u data: %d (compressed: %d)", p, planes[p].size(), res );
        }

        if ( bmhd->masking() == ChunkBMHD::MASK_HAS_MASK ) {
            size_t res = decode_row( current_offset,
                                     row_bytes,
                                     bmhd->compression(),
                                     mask );
            if ( res == 0 ) {
                LOG_CRITICAL( "invalid body" );
                return -EINVAL;
            }

            current_offset += res;

            LOG_DEBUG( "mask data: %d (compressed: %d)", mask.size(), res );
        }
    }

    if ( current_offset < m_raw_data.size() ) {
        LOG_DEBUG( "body data left: %u", m_raw_data.size() - current_offset );
    }

    return 0;
}

int ChunkBODY::planes_to_pixels( const std::shared_ptr< ChunkBMHD > &bmhd,
                                 const std::vector< std::vector< uint8_t > > &planes,
                                 const std::vector< uint8_t > &mask,
                                 std::vector< uint32_t > &res ) const
{
    const uint8_t nplanes = bmhd->nplanes();

    size_t pixels_stored = 0;

    if ( planes.size() < nplanes ) {
        LOG_CRITICAL( "not enough planes to convert" );
        return -EINVAL;
    }

    size_t pixels_per_line = bmhd->line_pitch();
    
    // tranpose nplanes * 8 bit into 8 pixel values (utilizing nplanes bits)
    for ( size_t offset = 0; offset < planes.at( 0 ).size(); offset++ ) {
        uint32_t pixels[8] = { 0 };
        uint8_t mask_pixel[8] = { 255, 255, 255, 255, 255, 255, 255, 255 };

        for ( uint8_t p = 0; p < nplanes; p++ ) {
            uint8_t v = 0;

            if ( offset < planes[ nplanes - 1 - p ].size() ) {
                v = planes[ nplanes - 1 - p ][ offset ];
            }

            for ( uint8_t bit = 0; bit < 8; bit++ ) {
                pixels[7 - bit] <<= 1;
                pixels[7 - bit] |= v & 1;
                v >>= 1;
            }
        }

        if ( offset < mask.size() ) {
            uint8_t mask_value = mask.at( offset );
            for ( uint8_t bit = 0; bit < 8; bit++ ) {
                if ( ! ( mask_value & 1 ) ) {
                    mask_pixel[7 - bit] = 0;
                }
                mask_value >>= 1;
            }
        }

        for ( uint8_t pixel = 0; pixel < 8; pixel++ ) {
            if ( pixels_stored < pixels_per_line * bmhd->height() ) {
                uint32_t pixel_value = pixels[pixel];

                if ( nplanes <= 8 && bmhd->masking() == ChunkBMHD::MASK_HAS_TRANSPARENCY_COLOR ) {
                    if ( pixel_value != bmhd->transparencyColor() ) {
                        pixel_value += 255UL << 24;
                    }
                } else {
                    pixel_value += (uint32_t)mask_pixel[pixel] << 24;
                }
                res.push_back( pixel_value );
                pixels_stored++;
            }
        }
    }

    return 0;
}

