/* chunk_cat.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_cat.hh"
#include "find_chunk.hh"
#include <sstream>

int ChunkCAT::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();

    if ( chunklen < 12 ) {
        return -EINVAL;
    }

    if ( m_chunk_type != CHUNKNAME( "CAT " ) ) {
        return -EINVAL;
    }

    LOG_DEBUG( "found CAT chunk" );
    LOG_DEBUG( "parse %u bytes", chunklen );

    const char *buf = &data[0];

    m_content_hint = betoh_u32( get_u32( buf, 0 ) );

    buf += 4;
    chunklen -= 4;

    while ( chunklen >= 8 ) {
        size_t inner_parsed_len = 0;

        auto elem = check_and_parse( buf, chunklen, inner_parsed_len );

        if ( ! elem ) {
            LOG_DEBUG( "Failed to parse CAT entry" );
            return -EINVAL;
        }

        if ( elem->getChunkName() != "FORM" ) {
            if ( fail_on_unsupported() ) {
                LOG_CRITICAL( "Invalid form %s in CAT chunk", elem->getChunkName().c_str() );
                return -EINVAL;
            } else {
                LOG_VERBOSE( "Invalid form %s in CAT chunk", elem->getChunkName().c_str() );
            }
        }

        m_entries.push_back( elem );

        buf += inner_parsed_len;
        chunklen -= inner_parsed_len;
    }

    return 0;
}

std::string ChunkCAT::dump_as_readable_text( const shared_properties_t &shared_props )
{
    std::string res;

    for ( auto &chunk : m_entries ) {
        res += chunk->dump_as_readable_text( shared_props );
    }

    return res;
}

json_t *ChunkCAT::dump_as_json( const shared_properties_t &shared_props,
                                bool include_raw ) const
{
    LOG_DEBUG("called");

    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "CAT " ) );
    json_object_set_new( jroot, "subtype", json_string( CHUNKTYPE2NAME( m_content_hint ).c_str() ) );

    auto jarray = json_array();

    for ( auto &chunk : m_entries ) {
        auto jentry = chunk->dump_as_json( shared_props, include_raw );

        json_array_append_new( jarray, jentry );
    }

    json_object_set_new( jroot, "content", jarray );
    
    return jroot;
}

int ChunkCAT::dump_as_binary( const shared_properties_t &shared_props,
                              std::list< IFFBinaryResult > &output,
                              bool include_raw ) const
{
    LOG_DEBUG("called");

    for ( auto &chunk : m_entries ) {
        int res = chunk->dump_as_binary( shared_props, output, include_raw );
        if ( res != 0 ) {
            return res;
        }
    }
    
    return 0;
}

int ChunkCAT::fill_info( IFFInfo &info ) const
{
    std::stringstream ss1;

    info.set_file_type( "CAT " + CHUNKTYPE2NAME( m_content_hint ) );
    std::string descr = "Catalog of ";
    std::string addinfo = "";

    int res = -ENOTSUP;

    for ( auto &chunk : m_entries ) {
        IFFInfo sinfo;
        if ( chunk->fill_info( sinfo ) == 0 ) {
            if ( ! sinfo.description().empty() ) {
                descr += sinfo.description() + ";";
            }
            if ( ! sinfo.additional_info().empty() ) {
                addinfo += sinfo.additional_info() + ";";
            }
            res = 0;
        }
    }

    info.set_description( descr );
    info.set_additional_info( addinfo );

    return res;
}
