/* chunk_chrs.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_chrs.hh"
#include <algorithm>
#include <iconv.h>
#include <cstring>

int ChunkCHRS::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();
    const char *buf = &data[0];

    if ( m_chunk_type != CHUNKNAME( "CHRS" ) ) {
        return -EINVAL;
    }

    m_raw_text.resize( chunklen );
    std::copy_n( buf, chunklen,
                 m_raw_text.begin() );

    LOG_DEBUG( "Parsed CHRS" );

    return 0;
}

const std::vector< char > ChunkCHRS::raw_text() const
{
    return m_raw_text;
}

std::string ChunkCHRS::text() const
{
    std::string res;

    res.resize( m_raw_text.size() );
    std::copy( m_raw_text.begin(),
               m_raw_text.end(),
               res.begin() );

    if ( attribute( "chrs.newline", 0 ) ) {
        res += "\n";
    }

    return res;
}

int ChunkCHRS::convert_control_sequences( const std::string &in,
                                          std::string &out )
{
    auto inch = in.begin();
    while ( inch != in.end() ) {
        uint8_t ch = (uint8_t)*inch;
        inch++;

        if ( ch == '\n' ) {
            out += ch;
        } else if ( ch == 0x1b ) {
            while ( inch != in.end() &&
                    *inch >= 0x20 && *inch <= 0x2f ) {
                inch++;
            }
            if ( inch == in.end() ) break;
            if ( *inch < 0x30 || *inch >= 0x7f ) {
                LOG_DEBUG( "Invalid escape sequence" );
            }
            inch++;
        } else if ( ch >= 0x20 && ch <= 0x7f ) {
            out += ch;
        } else if ( ch >= 0xa0 && ch <= 0xff ) {
            out += ch;
        } else if ( ch == 0x8e || ch == 0x8f ) {
            if ( inch != in.end() ) {
                if ( *inch < 0x20 || *inch >= 0x7f ) {
                    LOG_DEBUG( "Invalid alternative graphics char" );
                }
                inch++;
            }
        } else if ( ch == 0x9b ) {
            if ( inch != in.end() ) {
                if ( *inch < 0x20 || *inch >= 0x7f ) {
                    LOG_DEBUG( "Invalid CSI sequence" );
                    inch++;
                } else {
                    while ( inch != in.end() &&
                            *inch >= 0x20 && *inch <= 0x3f ) {
                        inch++;
                    }
                    if ( inch == in.end() ) break;
                    if ( *inch < 0x40 || *inch > 0x7e ) {
                        LOG_DEBUG( "Invalid CSI sequence" );
                    }
                    inch++;
                }
            }
        } else if ( ch == 0x90 || ch == 0x9d || ch == 0x9e || ch == 0x9f ) {
            while ( inch != in.end() &&
                    *inch >= 0x20 && *inch <= 0x3f ) {
                inch++;
            }
            if ( inch == in.end() ) break;
            if ( *inch == 0x9c || *inch >= 0xa0 ) {
            } else {
                LOG_DEBUG( "Invalid DCS sequence" );
            }
            inch++;
        } else {
            // discard
        }
    }

    return 0;
}

int ChunkCHRS::convert_to_utf8( const std::string &in,
                                std::string &out )
{
    iconv_t cd = iconv_open( "UTF-8", "LATIN1" );
    if ( cd != (iconv_t)-1 ) {
        char *outbuffer = (char*)malloc( 10 * in.size() );
        char *inbuffer = strdup( in.c_str() );
        if ( inbuffer && outbuffer ) {
            memset( outbuffer, 0, 10 * in.size() );
            char *output = outbuffer;
            char *input = inbuffer;
            size_t outsize = 10 * in.size();
            size_t insize = in.size();
            size_t prev_outsize = outsize;

            size_t iconv_res = iconv( cd,
                                      &input, &insize,
                                      &output, &outsize );

            if ( iconv_res != (size_t)-1 && insize == 0 ) {
                out = std::string( outbuffer , prev_outsize - outsize );
            } else {
                LOG_DEBUG( "Failed to convert text to utf8" );
            }
        } else {
            out.clear();
        }
        free( inbuffer );
        free( outbuffer );
        iconv_close( cd );
    } else {
        LOG_DEBUG( "Failed to create iconv context" );
        out.clear();
    }

    return 0;
}

std::string ChunkCHRS::parsed_text_utf8() const
{
    std::string in = text();
    std::string out;

    if ( convert_control_sequences( in, out ) != 0 ) {
        return "";
    }

    in = out;

    if ( convert_to_utf8( in, out ) != 0 ) {
        return "";
    }

    return out;
}

std::string ChunkCHRS::dump_as_readable_text( const shared_properties_t &shared_props )
{
    return parsed_text_utf8();
}

json_t *ChunkCHRS::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );

    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( CHUNKTYPE2NAME( m_chunk_type ).c_str() ) );

    json_object_set_new( jroot, "parsedText", json_string( parsed_text_utf8().c_str() ) );

    if ( include_raw ) {
        auto jarray = json_array();
        for ( auto &ch : m_raw_text ) {
            json_array_append_new( jarray, json_integer( (uint8_t)ch ) );
        }
        json_object_set_new( jroot, "rawText", jarray );
    }

    return jroot;
}

int ChunkCHRS::dump_as_raw( const shared_properties_t &shared_props,
                            std::vector< char > &data ) const
{
    LOG_DEBUG( "called" );

    data = m_raw_text;

    return 0;
}
