/* chunk_cmap.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_cmap.hh"

int ChunkCMAP::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();

    if ( m_chunk_type != CHUNKNAME( "CMAP" ) ) {
        return -EINVAL;
    }

    const char *buf = &data[0];

    if ( ( chunklen % 3 ) != 0 ) {
        LOG_DEBUG( "invalid CMAP chunk size: %u", chunklen );
        return -EINVAL;
    }

    size_t colors = chunklen / 3;
    m_colors.clear();

    while ( colors > 0 ) {
        m_colors.push_back( cmap_raw_color( buf[0],
                                            buf[1],
                                            buf[2] ) );
        buf += 3;
        colors--;
    }

    return 0;
}

json_t *ChunkCMAP::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "CMAP" ) );
    json_object_set_new( jroot, "ncolors", json_integer( m_colors.size() ) );

    auto jcolors = json_array();
    
    for ( const auto &c : m_colors ) {
        auto jcolor = json_object();
        json_object_set_new( jcolor, "red", json_integer( c.red ) );
        json_object_set_new( jcolor, "green", json_integer( c.green ) );
        json_object_set_new( jcolor, "blue", json_integer( c.blue ) );
        json_array_append_new( jcolors, jcolor );
    }

    json_object_set_new( jroot, "colors", jcolors );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

ChunkCMAP::cmap_raw_color ChunkCMAP::get_color( uint8_t palette_entry ) const
{
    if ( palette_entry >= m_colors.size() ) {
        return cmap_raw_color();
    }

    return m_colors.at( palette_entry );
}

size_t ChunkCMAP::colors() const
{
    return m_colors.size();
}
