/* chunk_crng.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_crng.hh"
#include <cmath>

int ChunkCRNG::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();

    if ( m_chunk_type != CHUNKNAME( "CRNG" ) ) {
        return -EINVAL;
    }

    const char *buf = &data[0];

    if ( chunklen != 8 ) {
        LOG_DEBUG( "invalid CRNG chunk size: %u", chunklen );
        return -EINVAL;
    }

    m_cycle_rate = betoh_u16( get_u16( buf, 2 ) );
    m_flags = betoh_u16( get_u16( buf, 4 ) );
    m_low_color = get_u8( buf, 6 );
    m_high_color = get_u8( buf, 7 );

    return 0;
}

json_t *ChunkCRNG::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( CHUNKTYPE2NAME( m_chunk_type ).c_str() ) );
    json_object_set_new( jroot, "cycleRateValue", json_integer( m_cycle_rate ) );
    json_object_set_new( jroot, "stepsPerSecond", json_integer( (int)round( (double)m_cycle_rate * 60 / 16384 ) ) );
    if ( m_cycle_rate != 0 ) {
        json_object_set_new( jroot, "millisecondsPerStep", json_integer( (int)round( 16384.0 * 1000.0 / ( m_cycle_rate * 60 ) ) ) );
    }
    json_object_set_new( jroot, "flagsValue", json_integer( m_flags ) );
    json_object_set_new( jroot, "lowColor", json_integer( m_low_color ) );
    json_object_set_new( jroot, "highColor", json_integer( m_high_color ) );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}
