/* chunk_dbod.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_dbod.hh"
#include "chunk_dpel.hh"
#include "chunk_dgbl.hh"
#include "chunk_dloc.hh"

int ChunkDBOD::parse_chunk( const std::vector< char > &data )
{
    //auto chunklen = data.size();

    if ( m_chunk_type != CHUNKNAME( "DBOD" ) ) {
        return -EINVAL;
    }

    return 0;
}

json_t *ChunkDBOD::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "DBOD" ) );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

int ChunkDBOD::decodeRGBA( const std::shared_ptr< ChunkDGBL > &dgbl,
                           const std::shared_ptr< ChunkDPEL > &dpel,
                           const std::shared_ptr< ChunkDLOC > &dloc,
                           std::vector< uint32_t > &res ) const
{
    struct pixel_element {
        bool use = false;
        uint8_t type = 0;
        uint8_t length = 0;
        uint32_t mask = 0;
        uint8_t shift = 0;
    };

    if ( dgbl->compression() != ChunkDGBL::COMPRESSION_NONE ) {
        LOG_CRITICAL( "Compression not supported" );
        return -ENOTSUP;
    }

    std::vector< pixel_element > pes;

    auto elements = dpel->elements();

    for ( auto &e : elements ) {
        pixel_element pe;

        pe.length = ( e.depth + 7 ) / 8;

        if ( e.type == ChunkDPEL::TYPE_RED ) {
            pe.mask = 0xff;
            pe.shift = 0;
            pe.type = e.type;
            pe.use = true;
        } else if ( e.type == ChunkDPEL::TYPE_GREEN ) {
            pe.mask = 0xff;
            pe.shift = 8;
            pe.type = e.type;
            pe.use = true;
        } else if ( e.type == ChunkDPEL::TYPE_BLUE ) {
            pe.mask = 0xff;
            pe.shift = 16;
            pe.type = e.type;
            pe.use = true;
        } else if ( e.type == ChunkDPEL::TYPE_ALPHA ) {
            pe.mask = 0xff;
            pe.shift = 24;
            pe.type = e.type;
            pe.use = true;
        } else {
            LOG_VERBOSE( "channels other than RGBA are not supported" );
        }
        pes.push_back( pe );
    }

    size_t offset = 0;

    while ( offset < m_raw_data.size() ) {
        uint32_t pixel = 0;
        for ( auto &pe : pes ) {
            auto e = get_pixel_data( offset, pe.length );
            if ( ! e.second ) {
                break;
            }

            offset += pe.length;

            if ( pe.use ) {
                uint32_t v = e.first & pe.mask;
                v <<= pe.shift;
                pixel |= v;
            }
        }

        LOG_DEBUG( "got pixel value %x", pixel );

        res.push_back( pixel );
    }

    return 0;
}

std::pair< uint32_t, bool > ChunkDBOD::get_pixel_data( size_t offset, uint8_t length ) const
{
    if ( offset + length > m_raw_data.size() ) {
        return std::make_pair( 0, false );
    }

    uint32_t v = 0;
    while ( length > 0 ) {
        v <<= 8;
        v |= m_raw_data.at( offset );
        offset++;
        length--;
    }

    return std::make_pair( v, true );
}
