/* chunk_dlta.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_dlta.hh"
#include "chunk_bmhd.hh"
#include "chunk_anhd.hh"
#include "util.hh"

int ChunkDLTA::parse_chunk( const std::vector< char > &data )
{
    if ( m_chunk_type != CHUNKNAME( "DLTA" ) ) {
        return -EINVAL;
    }

    return 0;
}

json_t *ChunkDLTA::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( CHUNKTYPE2NAME( m_chunk_type ).c_str() ) );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

int ChunkDLTA::apply_delta( const std::shared_ptr< ChunkBMHD > &bmhd,
                            const std::shared_ptr< ChunkANHD > &anhd,
                            const std::vector< std::vector< uint8_t > > &planes,
                            std::vector< std::vector< uint8_t > > &res ) const
{
    if ( anhd->operation() == ChunkANHD::OPERATION_VERTICAL_DELTA ) {
        return apply_delta5( bmhd, anhd,
                             planes, res );
    } else if ( anhd->operation() == ChunkANHD::OPERATION_SHORT_LONG_VDELTA ) {
        return apply_delta7( bmhd, anhd,
                             planes, res );
    } else if ( anhd->operation() == ChunkANHD::OPERATION_DELTA8 ) {
        return apply_delta8( bmhd, anhd,
                             planes, res );
    } else if ( anhd->operation() == ChunkANHD::OPERATION_LONG_DELTA ) {
        return apply_delta23( bmhd, anhd,
                              planes, res, true );
    } else if ( anhd->operation() == ChunkANHD::OPERATION_SHORT_DELTA ) {
        return apply_delta23( bmhd, anhd,
                              planes, res, false );
    } else if ( anhd->operation() == ChunkANHD::OPERATION_SHORT_LONG_DELTA ) {
        return apply_delta4( bmhd, anhd,
                             planes, res );
    }

    LOG_CRITICAL( "delta operation %u not supported", anhd->operation() );

    return -ENOTSUP;
}

int ChunkDLTA::apply_delta5( const std::shared_ptr< ChunkBMHD > &bmhd,
                             const std::shared_ptr< ChunkANHD > &anhd,
                             const std::vector< std::vector< uint8_t > > &planes,
                             std::vector< std::vector< uint8_t > > &res ) const
{
    if ( ( anhd->bits() & ~( 1 << 1 ) ) != 0 ) {
        LOG_CRITICAL( "option bits not supported" );
    }

    bool xor_mode = anhd->bits() & ( 1 << 1 );

    size_t cols = ( bmhd->width() + 7 ) / 8;
    size_t pitch = ( ( bmhd->width() + 15 ) / 16 ) * 2;
    size_t current_offset = 0;
    const uint8_t *data = (const uint8_t *)&m_raw_data[0];

    if ( m_raw_data.size() < 64 ) {
        LOG_CRITICAL( "DLTA chunk too small" );
        return -EINVAL;
    }

    for ( uint8_t plane = 0; plane < bmhd->nplanes(); plane++ ) {
        uint32_t plane_delta_ptr = be32toh( ((const uint32_t *)data)[ plane ] );
        LOG_DEBUG( "got data pointer %u for plane %u", plane_delta_ptr, plane );

        auto plane_data = planes[ plane ];

        if ( plane_delta_ptr == 0 ) {
            res.push_back( plane_data );
            continue;
        }

        if ( plane_delta_ptr >= m_raw_data.size() ) {
            LOG_CRITICAL( "DLTA chunk too small" );
            return -EINVAL;
        }

        current_offset = plane_delta_ptr;

        for ( uint16_t x = 0; x < cols; x++ ) {
            if ( current_offset >= m_raw_data.size() ) {
                LOG_CRITICAL( "DLTA chunk too small" );
                return -EINVAL;
            }

            uint8_t no_of_ops = data[ current_offset++ ];
            uint8_t *column_data = &plane_data[x];
            uint8_t *column_end = column_data + bmhd->height() * pitch;

            LOG_DEBUG( "got %u ops for column %u", no_of_ops, x );

            for ( ; no_of_ops > 0; no_of_ops-- ) {
                if ( current_offset >= m_raw_data.size() ) {
                    LOG_CRITICAL( "DLTA chunk too small" );
                    return -EINVAL;
                }

                const uint8_t op = data[ current_offset++ ];

                if ( op & 0x80 ) {
                    // copy data
                    LOG_DEBUG( "copy %u bytes", op & 0x7f );

                    for ( uint8_t bytes = 0; bytes < ( op & 0x7f ); bytes++ ) {
                        if ( current_offset >= m_raw_data.size() ) {
                            LOG_CRITICAL( "DLTA chunk too small" );
                            return -EINVAL;
                        }

                        if ( column_data >= column_end ) {
                            LOG_CRITICAL( "more ops than rows, ignoring remaining" );
                        } else {
                            const uint8_t b = data[ current_offset++ ];
                            LOG_DEBUG( "  copy byte %u", b );

                            if ( xor_mode ) {
                                *column_data ^= b;
                            } else {
                                *column_data = b;
                            }
                        }
                        column_data += pitch;
                    }
                } else if ( op == 0 ) {
                    // duplicate byte

                    if ( current_offset + 1 >= m_raw_data.size() ) {
                        LOG_CRITICAL( "DLTA chunk too small" );
                        return -EINVAL;
                    }

                    uint8_t count = data[ current_offset++ ];
                    const uint8_t value = data[ current_offset++ ];

                    LOG_DEBUG( "duplicate byte %u %u times", value, count );

                    for ( ; count > 0; count-- ) {
                        if ( column_data >= column_end ) {
                            LOG_CRITICAL( "more ops than rows, ignoring remaining" );
                        } else {
                            if ( xor_mode ) {
                                *column_data ^= value;
                            } else {
                                *column_data = value;
                            }
                        }
                        column_data += pitch;
                    }
                } else {
                    // skip
                    LOG_DEBUG( "skip %u rows", op );

                    column_data += pitch * op;
                }
            }
        }

        res.push_back( plane_data );
    }

    return 0;
}

int ChunkDLTA::apply_delta7( const std::shared_ptr< ChunkBMHD > &bmhd,
                             const std::shared_ptr< ChunkANHD > &anhd,
                             const std::vector< std::vector< uint8_t > > &planes,
                             std::vector< std::vector< uint8_t > > &res ) const
{
    if ( ( anhd->bits() & ~( 1 << 0 ) ) != 0 ) {
        LOG_CRITICAL( "option bits not supported" );
    }

    bool long_data = anhd->bits() & ( 1 << 0 );

    size_t cols = ( bmhd->width() + 15 ) / 16;
    if ( long_data ) {
        cols = ( bmhd->width() + 31 ) / 32;
    }

    size_t pitch = ( ( bmhd->width() + 15 ) / 16 ) * 2;
    size_t current_offset = 0;
    size_t current_data_offset = 0;
    const uint8_t *data = (const uint8_t *)&m_raw_data[0];

    if ( m_raw_data.size() < 64 ) {
        LOG_CRITICAL( "DLTA chunk too small" );
        return -EINVAL;
    }

    for ( uint8_t plane = 0; plane < bmhd->nplanes(); plane++ ) {
        uint32_t ops_ptr = be32toh( ((const uint32_t *)data)[ plane ] );
        uint32_t data_ptr = be32toh( ((const uint32_t *)data)[ plane + 8 ] );
        LOG_DEBUG( "got ops and data pointer %u, %u for plane %u", ops_ptr, data_ptr, plane );

        auto plane_data = planes[ plane ];
        auto plane_data_start = &plane_data[0];
        auto plane_data_end = plane_data_start + plane_data.size();

        if ( ops_ptr == 0 ) {
            res.push_back( plane_data );
            continue;
        }

        if ( ops_ptr >= m_raw_data.size() ) {
            LOG_CRITICAL( "DLTA chunk too small" );
            return -EINVAL;
        }

        current_offset = ops_ptr;
        current_data_offset = data_ptr;

        for ( uint16_t x = 0; x < cols; x++ ) {
            if ( current_offset >= m_raw_data.size() ) {
                LOG_CRITICAL( "DLTA chunk too small" );
                return -EINVAL;
            }

            uint8_t no_of_ops = data[ current_offset++ ];
            uint8_t *column_data = &plane_data[x * ( long_data ? 4 : 2 )];
            uint8_t *column_end = column_data + bmhd->height() * pitch;
            
            LOG_DEBUG( "got %u ops for column %u", no_of_ops, x );

            for ( ; no_of_ops > 0; no_of_ops-- ) {
                if ( current_offset >= m_raw_data.size() ) {
                    LOG_CRITICAL( "DLTA chunk too small" );
                    return -EINVAL;
                }

                const uint8_t op = data[ current_offset++ ];

                if ( op & 0x80 ) {
                    // copy data
                    LOG_DEBUG( "copy %u elements", op & 0x7f );

                    for ( uint8_t elements = 0; elements < ( op & 0x7f ); elements++ ) {
                        if ( current_data_offset + ( long_data ? 3 : 1 ) >= m_raw_data.size() ) {
                            LOG_CRITICAL( "DLTA chunk too small" );
                            return -EINVAL;
                        }

                        if ( column_data >= column_end ) {
                            LOG_CRITICAL( "more ops than rows, ignoring remaining" );
                        } else {
                            if ( column_data < plane_data_start ||
                                 column_data + ( long_data ? 4 : 2 ) > plane_data_end ) {
                                LOG_CRITICAL( "invalid pointer" );
                                return -EINVAL;
                            }

                            uint8_t s = a_min( long_data ? 4 : 2, pitch - x * ( long_data ? 4 : 2 ) );

                            memcpy( column_data,
                                    &data[ current_data_offset ],
                                    s );

                            current_data_offset += long_data ? 4 : 2;
                        }
                        column_data += pitch;
                    }
                } else if ( op == 0 ) {
                    // duplicate byte

                    if ( current_offset + 1 >= m_raw_data.size() ) {
                        LOG_CRITICAL( "DLTA chunk too small" );
                        return -EINVAL;
                    }

                    uint8_t count = data[ current_offset++ ];

                    if ( current_data_offset + ( long_data ? 3 : 1 ) >= m_raw_data.size() ) {
                        LOG_CRITICAL( "DLTA chunk too small" );
                        return -EINVAL;
                    }

                    LOG_DEBUG( "duplicate data %u times", count );

                    for ( ; count > 0; count-- ) {
                        if ( column_data >= column_end ) {
                            LOG_CRITICAL( "more ops than rows, ignoring remaining" );
                        } else {
                            if ( column_data < plane_data_start ||
                                 column_data + 4 > plane_data_end ) {
                                LOG_CRITICAL( "invalid pointer" );
                                return -EINVAL;
                            }

                            uint8_t s = a_min( long_data ? 4 : 2, pitch - x * ( long_data ? 4 : 2 ) );

                            memcpy( column_data,
                                    &data[ current_data_offset ],
                                    s );
                        }
                        column_data += pitch;
                    }

                    current_data_offset += long_data ? 4 : 2;
                } else {
                    // skip
                    LOG_DEBUG( "skip %u rows", op );

                    column_data += pitch * op;
                }
            }
        }

        res.push_back( plane_data );
    }

    return 0;
}

int ChunkDLTA::apply_delta8( const std::shared_ptr< ChunkBMHD > &bmhd,
                             const std::shared_ptr< ChunkANHD > &anhd,
                             const std::vector< std::vector< uint8_t > > &planes,
                             std::vector< std::vector< uint8_t > > &res ) const
{
    if ( ( anhd->bits() & ~( 1 << 0 | 1 << 1 ) ) != 0 ) {
        LOG_CRITICAL( "option bits not supported" );
    }

    bool long_data = anhd->bits() & ( 1 << 0 );
    bool xor_mode = anhd->bits() & ( 1 << 1 );

    size_t cols = ( bmhd->width() + 15 ) / 16;
    if ( long_data ) {
        cols = ( bmhd->width() + 31 ) / 32;
    }

    size_t pitch = ( ( bmhd->width() + 15 ) / 16 ) * 2;
    size_t current_offset = 0;
    const uint8_t *data = (const uint8_t *)&m_raw_data[0];

    if ( m_raw_data.size() < 64 ) {
        LOG_CRITICAL( "DLTA chunk too small" );
        return -EINVAL;
    }

    for ( uint8_t plane = 0; plane < bmhd->nplanes(); plane++ ) {
        uint32_t plane_delta_ptr = be32toh( ((const uint32_t *)data)[ plane ] );
        LOG_DEBUG( "got data pointer %u for plane %u", plane_delta_ptr, plane );

        auto plane_data = planes[ plane ];

        if ( plane_delta_ptr == 0 ) {
            res.push_back( plane_data );
            continue;
        }

        if ( plane_delta_ptr >= m_raw_data.size() ) {
            LOG_CRITICAL( "DLTA chunk too small" );
            return -EINVAL;
        }

        current_offset = plane_delta_ptr;

        for ( uint16_t x = 0; x < cols; x++ ) {
            if ( current_offset >= m_raw_data.size() ) {
                LOG_CRITICAL( "DLTA chunk too small" );
                return -EINVAL;
            }

            uint32_t no_of_ops;
            if ( long_data ) {
                no_of_ops = be32toh( get_u32( data,  current_offset ) );
                current_offset += 4;
            } else {
                no_of_ops = be16toh( get_u16( data, current_offset ) );
                current_offset += 2;
            }
            uint8_t *column_data = &plane_data[x * ( long_data ? 4 : 2 )];
            uint8_t *column_end = column_data + bmhd->height() * pitch;

            LOG_DEBUG( "got %u ops for column %u", no_of_ops, x );

            for ( ; no_of_ops > 0; no_of_ops-- ) {
                if ( current_offset >= m_raw_data.size() ) {
                    LOG_CRITICAL( "DLTA chunk too small" );
                    return -EINVAL;
                }

                uint32_t op;
                if ( long_data ) {
                    op = be32toh( get_u32( data, current_offset  ) );
                    current_offset += 4;
                } else {
                    op = be16toh( get_u16( data, current_offset ) );
                    current_offset += 2;
                    op <<= 16;
                }

                if ( op & 0x80000000 ) {
                    // copy data
                    uint32_t count = op & 0x7fffffff;
                    if ( ! long_data ) {
                        count >>= 16;
                    }
                    LOG_DEBUG( "copy %u bytes", count );

                    for ( uint8_t elements = 0; elements < count; elements++ ) {
                        if ( current_offset + ( long_data ? 4 : 2 ) > m_raw_data.size() ) {
                            LOG_CRITICAL( "DLTA chunk too small" );
                            return -EINVAL;
                        }

                        if ( column_data >= column_end ) {
                            LOG_CRITICAL( "more ops than rows, ignoring remaining" );
                        } else {
                            if ( long_data ) {
                                const uint32_t d = *( (uint32_t*)( &data[ current_offset ] ) );
                                uint32_t *target = (uint32_t*)column_data;

                                if ( xor_mode ) {
                                    *target ^= d;
                                } else {
                                    *target = d;
                                }

                                current_offset += 4;
                            } else {
                                const uint16_t d = *( (uint16_t*)( &data[ current_offset ] ) );
                                uint16_t *target = (uint16_t*)column_data;

                                if ( xor_mode ) {
                                    *target ^= d;
                                } else {
                                    *target = d;
                                }

                                current_offset += 2;
                            }
                        }
                        column_data += pitch;
                    }
                } else if ( op == 0 ) {
                    // duplicate byte

                    if ( current_offset + 2 * ( long_data ? 4 : 2 ) > m_raw_data.size() ) {
                        LOG_CRITICAL( "DLTA chunk too small" );
                        return -EINVAL;
                    }

                    uint32_t count;
                    if ( long_data ) {
                        count = be32toh( get_u32( data, current_offset ) );
                        current_offset += 4;
                    } else {
                        count = be16toh( get_u16( data, current_offset ) );
                        current_offset += 2;
                    }
                    const uint8_t *source = &data[ current_offset ];
                    current_offset += long_data ? 4 : 2;

                    LOG_DEBUG( "duplicate byte %u times", count );

                    for ( ; count > 0; count-- ) {
                        if ( column_data >= column_end ) {
                            LOG_CRITICAL( "more ops than rows, ignoring remaining" );
                        } else {
                            for ( uint8_t b = 0; b < ( long_data ? 4 : 2 ); b++ ) {
                                if ( xor_mode ) {
                                    column_data[b] ^= source[b];
                                } else {
                                    column_data[b] = source[b];
                                }
                            }
                        }
                        column_data += pitch;
                    }
                } else {
                    // skip
                    if ( ! long_data ) {
                        op >>= 16;
                    }

                    LOG_DEBUG( "skip %u rows", op );

                    column_data += pitch * op;
                }
            }
        }

        res.push_back( plane_data );
    }

    return 0;
}

int ChunkDLTA::apply_delta23( const std::shared_ptr< ChunkBMHD > &bmhd,
                              const std::shared_ptr< ChunkANHD > &anhd,
                              const std::vector< std::vector< uint8_t > > &planes,
                              std::vector< std::vector< uint8_t > > &res,
                              bool long_data ) const
{
    // this code is untested!

    size_t pitch = ( ( bmhd->width() + 15 ) / 16 ) * 2;
    size_t current_offset = 0;
    const uint8_t *data = (const uint8_t *)&m_raw_data[0];

    if ( m_raw_data.size() < 32 ) {
        LOG_CRITICAL( "DLTA chunk too small" );
        return -EINVAL;
    }

    for ( uint8_t plane = 0; plane < bmhd->nplanes(); plane++ ) {
        uint32_t plane_delta_ptr = be32toh( ((const uint32_t *)data)[ plane ] );
        LOG_DEBUG( "got data pointer %u for plane %u", plane_delta_ptr, plane );

        auto plane_data = planes[ plane ];

        if ( plane_delta_ptr == 0 ) {
            res.push_back( plane_data );
            continue;
        }

        if ( plane_delta_ptr >= m_raw_data.size() ) {
            LOG_CRITICAL( "DLTA chunk too small" );
            return -EINVAL;
        }

        current_offset = plane_delta_ptr;

        uint8_t *target = &plane_data[0];
        uint8_t *target_end = target + bmhd->height() * pitch;

        while ( current_offset + 1 < m_raw_data.size() ) {
            int16_t op = (int16_t)be16toh( get_u16( data, current_offset ) );
            current_offset += 2;

            if ( op == 0xffff ) {
                break;
            }

            if ( op >= 0 ) {
                target = &plane_data[0];
                target += ( long_data ? 4 : 2 ) * op;

                if ( current_offset + ( long_data ? 4 : 2 ) > m_raw_data.size() ) {
                    LOG_CRITICAL( "DLTA chunk too small" );
                    return -EINVAL;
                }

                if ( target + ( long_data ? 4 : 2 ) > target_end ) {
                    LOG_CRITICAL( "plane data overflow" );
                    return -EINVAL;
                }

                memcpy( target, &data[ current_offset ],
                        long_data ? 4 : 2 );

                current_offset += long_data ? 4 : 2;
            } else {
                target = &plane_data[0];
                target += ( long_data ? 4 : 2 ) * -( op + 2 );

                if ( current_offset + ( long_data ? 4 : 2 ) > m_raw_data.size() ) {
                    LOG_CRITICAL( "DLTA chunk too small" );
                    return -EINVAL;
                }

                uint16_t words = be16toh( get_u16( data, current_offset ) );
                current_offset += 2;

                if ( current_offset + words * ( long_data ? 4 : 2 ) > m_raw_data.size() ) {
                    LOG_CRITICAL( "DLTA chunk too small" );
                    return -EINVAL;
                }

                if ( target + words * ( long_data ? 4 : 2 ) > target_end ) {
                    LOG_CRITICAL( "plane data overflow" );
                    return -EINVAL;
                }

                memcpy( target, &data[ current_offset ],
                        words * ( long_data ? 4 : 2 ) );

                current_offset += words * ( long_data ? 4 : 2 );
            }
        }

        res.push_back( plane_data );
    }

    return 0;
}

int ChunkDLTA::apply_delta4( const std::shared_ptr< ChunkBMHD > &bmhd,
                             const std::shared_ptr< ChunkANHD > &anhd,
                             const std::vector< std::vector< uint8_t > > &planes,
                             std::vector< std::vector< uint8_t > > &res ) const
{
    // this code is untested!

    bool long_data = anhd->bits() & ( 1 << 0 );
    size_t pitch = ( ( bmhd->width() + 15 ) / 16 ) * 2;
    size_t current_offset = 0;
    const uint8_t *data = (const uint8_t *)&m_raw_data[0];

    if ( m_raw_data.size() < 64 ) {
        LOG_CRITICAL( "DLTA chunk too small" );
        return -EINVAL;
    }

    for ( uint8_t plane = 0; plane < bmhd->nplanes(); plane++ ) {
        uint32_t data_ptr = be32toh( ((const uint32_t *)data)[ plane ] );
        uint32_t ops_ptr = be32toh( ((const uint32_t *)data)[ plane + 8 ] );

        auto plane_data = planes[ plane ];

        if ( ops_ptr == 0 ) {
            res.push_back( plane_data );
            continue;
        }

        if ( ops_ptr >= m_raw_data.size() ) {
            LOG_CRITICAL( "DLTA chunk too small" );
            return -EINVAL;
        }

        current_offset = ops_ptr;

        uint8_t *target = &plane_data[0];
        uint8_t *target_end = target + bmhd->height() * pitch;

        while ( current_offset + 1 < m_raw_data.size() ) {
            uint16_t op = be16toh( get_u16( data, current_offset ) );
            current_offset += 2;

            if ( op == 0xffff ) {
                break;
            }

            target = &plane_data[0];
            target += ( long_data ? 4 : 2 ) * op;

            if ( current_offset + ( long_data ? 4 : 2 ) > m_raw_data.size() ) {
                LOG_CRITICAL( "DLTA chunk too small" );
                return -EINVAL;
            }

            int16_t size = (int16_t)be16toh( get_u16( data, current_offset ) );
            current_offset += 2;

            if ( size >= 0 ) {
                if ( data_ptr + ( long_data ? 4 : 2 ) * size > m_raw_data.size() ) {
                    LOG_CRITICAL( "DLTA chunk too small" );
                    return -EINVAL;
                }

                if ( target + ( long_data ? 4 : 2 ) * size > target_end ) {
                    LOG_CRITICAL( "plane data overflow" );
                    return -EINVAL;
                }

                memcpy( target, &data[ data_ptr ],
                        ( long_data ? 4 : 2 ) * size );

                data_ptr += ( long_data ? 4 : 2 ) * size;
            } else {
                if ( data_ptr + ( long_data ? 4 : 2 ) > m_raw_data.size() ) {
                    LOG_CRITICAL( "DLTA chunk too small" );
                    return -EINVAL;
                }

                if ( target + ( long_data ? 4 : 2 ) * size > target_end ) {
                    LOG_CRITICAL( "plane data overflow" );
                    return -EINVAL;
                }

                for ( int16_t s = size; s < 0; s++ ) {
                    memcpy( target, &data[ data_ptr ],
                            ( long_data ? 4 : 2 ) );

                    target += ( long_data ? 4 : 2 );
                }

                data_ptr += ( long_data ? 4 : 2 );
            }
        }

        res.push_back( plane_data );
    }

    return 0;
}
