/* chunk_dpel.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_dpel.hh"

int ChunkDPEL::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();

    LOG_DEBUG( "called" );

    if ( m_chunk_type != CHUNKNAME( "DPEL" ) ) {
        return -EINVAL;
    }

    const char *buf = &data[0];

    if ( chunklen < 4 ) {
        LOG_DEBUG( "invalid DPEL chunk size: %u", chunklen );
        return -EINVAL;
    }

    uint32_t elements = betoh_u32( get_u32( buf, 0 ) );

    if ( (size_t)( elements * 4 + 4 ) != chunklen ) {
        LOG_DEBUG( "invalid DPEL chunk size: %u/%u", elements, chunklen );
        return -EINVAL;
    }

    buf += 4;

    while ( elements > 0 ) {
        pixel_depth pd;
        pd.type = betoh_u16( get_u16( buf, 0 ) );
        pd.depth = betoh_u16( get_u16( buf, 2 ) );

        m_elements.push_back( pd );
        
        buf += 4;
        elements--;
    }

    return 0;
}

json_t *ChunkDPEL::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( CHUNKTYPE2NAME( m_chunk_type ).c_str() ) );
    json_object_set_new( jroot, "nelements", json_integer( m_elements.size() ) );

    auto jelements = json_array();
    for ( auto &e : m_elements ) {
        auto jelement = json_object();
        json_object_set_new( jelement, "type", json_integer( e.type ) );
        json_object_set_new( jelement, "depth", json_integer( e.depth ) );
        json_array_append_new( jelements, jelement );
    }
    json_object_set_new( jroot, "elements", jelements );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

std::vector< ChunkDPEL::pixel_depth > ChunkDPEL::elements() const
{
    return m_elements;
}
