/* chunk_fhdr.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_fhdr.hh"
#include <sstream>

int ChunkFHDR::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();

    if ( m_chunk_type != CHUNKNAME( "FHDR" ) ) {
        return -EINVAL;
    }

    const char *buf = &data[0];

    // 92 is just a guess
    if ( chunklen != 92 ) {
        LOG_DEBUG( "invalid FHDR chunk size: %u", chunklen );
        return -EINVAL;
    }

    m_depth = betoh_u16( get_u16( buf, 4 ) );
    m_width = betoh_u16( get_u16( buf, 6 ) );
    m_height = betoh_u16( get_u16( buf, 8 ) );

    for ( uint8_t c = 0; c < 32; c++ ) {
        uint16_t v = betoh_u16( get_u16( buf, 12 + c * 2 ) );
        uint8_t r = ( v >> 8 ) & 0xf;
        uint8_t g = ( v >> 4 ) & 0xf;
        uint8_t b = v & 0xf;
        ChunkCMAP::cmap_raw_color col( r << 4 | r,
                                       g << 4 | g,
                                       b << 4 | b );
        m_palette.push_back( col );
    }

    return 0;
}

json_t *ChunkFHDR::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "FHDR" ) );
    json_object_set_new( jroot, "width", json_integer( m_width ) );
    json_object_set_new( jroot, "height", json_integer( m_height ) );

    json_object_set_new( jroot, "depthValue", json_integer( m_depth ) );
    switch ( m_depth ) {
        case 1:
            json_object_set_new( jroot, "depth", json_string( "1" ) );
            break;
        case 2:
            json_object_set_new( jroot, "depth", json_string( "2" ) );
            break;
        case 3:
            json_object_set_new( jroot, "depth", json_string( "3" ) );
            break;
        case 4:
            json_object_set_new( jroot, "depth", json_string( "4" ) );
            break;
        case 5:
            json_object_set_new( jroot, "depth", json_string( "5" ) );
            break;
        case 6:
            json_object_set_new( jroot, "depth", json_string( "ham" ) );
            break;
        default:
            json_object_set_new( jroot, "depth", json_string( "invalid" ) );
            break;
    }

    auto palette = json_array();
    uint8_t col_id = 0;
    for ( auto col : m_palette ) {
        auto jcol = json_object();

        json_object_set_new( jcol, "id", json_integer( col_id++ ) );
        json_object_set_new( jcol, "red", json_integer( col.red ) );
        json_object_set_new( jcol, "green", json_integer( col.green ) );
        json_object_set_new( jcol, "blue", json_integer( col.blue ) );

        json_array_append_new( palette, jcol );
    }
    json_object_set_new( jroot, "palette", palette );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

uint16_t ChunkFHDR::width() const
{
    return m_width;
}

uint16_t ChunkFHDR::height() const
{
    return m_height;
}

uint16_t ChunkFHDR::depth() const
{
    return m_depth;
}

std::shared_ptr< ChunkCMAP > ChunkFHDR::get_cmap() const
{
    char buf[ 8 + 3 * m_palette.size() ];

    memcpy( buf, "CMAP", 4 );
    uint32_t *u32 = (uint32_t*)&buf[4];
    *u32 = htobe_u32( 3 * m_palette.size() );

    size_t offset = 8;
    for ( auto col : m_palette ) {
        buf[ offset++ ] = col.red;
        buf[ offset++ ] = col.green;
        buf[ offset++ ] = col.blue;
    }
    
    auto cmap = std::make_shared< ChunkCMAP >();
    size_t parsed_len;
    if ( cmap->parse( buf, sizeof( buf ), parsed_len ) != 0 ) {
        LOG_CRITICAL( "could not parse cmap" );
        return nullptr;
    }

    return cmap;
}

int ChunkFHDR::fill_info( IFFInfo &info ) const
{
    std::stringstream ss1;

    ss1 << width() << "x" << height();

    if ( m_depth == 6 ) {
        ss1 << " (" << ( 1 << 12 ) << " colors (HAM))";
    } else {
        ss1 << " (" << ( 1 << m_depth ) << " colors)";
    }

    info.set_additional_info( ss1.str() );

    return 0;
}
