/* chunk_form.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_form.hh"
#include "chunk_fons.hh"
#include "chunk_chrs.hh"
#include "chunk_wtxt.hh"
#include "chunk_wpar.hh"
#include "chunk_wowo_form.hh"
#include "find_chunk.hh"

// form ilbm
#include "chunk_bmhd.hh"
#include "chunk_cmap.hh"
#include "chunk_grab.hh"
#include "chunk_dest.hh"
#include "chunk_sprt.hh"
#include "chunk_camg.hh"
#include "chunk_body.hh"
#include "chunk_anno.hh"
#include "chunk_dpi.hh"
#include "chunk_crng.hh"

// form deep
#include "chunk_dgbl.hh"
#include "chunk_dloc.hh"
#include "chunk_dpel.hh"
#include "chunk_dchg.hh"
#include "chunk_dbod.hh"

#include "chunk_cstr.hh"

#include "chunk_name.hh"
#include "chunk_vhdr.hh"
#include "chunk_8svx_body.hh"

// form ilbm in anims
#include "chunk_anhd.hh"
#include "chunk_dlta.hh"

// ctlg
#include "chunk_fver.hh"
#include "chunk_lang.hh"

// gxui
#include "chunk_gxui_form.hh"

// fant
#include "chunk_fhdr.hh"

#include <sndfile.h>

#include <sstream>
#include <algorithm>

#include "util.hh"

const std::list< std::pair< std::string, std::list< std::string > > > ChunkFORM::m_list_of_known_chunks =
    {
     { "CPFM", { "IFHD",
                 "REFP",
                 "CHDT",
                 "CSNM" } },
     { "JBOB", { "BHDR",
                 "BODY" } },
     { "OPUS", { "TYPE",
                 "FUNC",
                 "ENVR",
                 "SETS",
                 "TBAR",
                 "LMEN",
                 "UMEN",
                 "SCRP",
                 "HKEY",
                 "BTNW",
                 "BUTN",
                 "ICON",
                 "LSTR",
                 "BANK" } },
     { "GXGA", { "GADA" } },
     { "GXBX", { "BBOX" } },
     { "GXTX", { "ITXT" } },
     { "GXMN", { "MEDA" } },
     { "FDDB", { "FDFH",
                 "FDCO",
                 "FDDR" } },
     { "FDPF", { "STAR",
                 "ASCI",
                 "SCOL",
                 "RPTP",
                 "PRTP",
                 "LBLP",
                 "MACH",
                 "MACD" } }
     };

int ChunkFORM::parseFTXT( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "FTXT" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "FONS" ) ) {
            auto fons = std::make_shared< ChunkFONS >();
            if ( fons->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ftxt_chunks.push_back( fons );
        } else if ( chunktype == CHUNKNAME( "CHRS" ) ) {
            LOG_DEBUG( "Got CHRS chunk of %u bytes", chunklen );
            auto chrs = std::make_shared< ChunkCHRS >();
            if ( chrs->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ftxt_chunks.push_back( chrs );

            //std::cout << chrs->parsed_text_utf8() << std::endl;
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "TCPP",
                                                                                        "KWVR",
                                                                                        "TCPH",
                                                                                        "TCHF",
                                                                                        "TCFF" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_ftxt_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "Formatted text" );

    return 0;
}

int ChunkFORM::parseWOWO( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "WOWO" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "WTXT" ) ) {
            LOG_DEBUG( "Got WTXT chunk of %u bytes", chunklen );
            auto chrs = std::make_shared< ChunkWTXT >();
            if ( chrs->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_wowo_chunks.push_back( chrs );
        } else if ( chunktype == CHUNKNAME( "WPAR" ) ) {
            LOG_DEBUG( "Got WPAR chunk of %u bytes", chunklen );
            auto wpar = std::make_shared< ChunkWPAR >();
            if ( wpar->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_wowo_chunks.push_back( wpar );
        } else if ( chunktype == CHUNKNAME( "FORM" ) ) {
            LOG_DEBUG( "Got FORM chunk of %u bytes", chunklen );
            auto form = std::make_shared< ChunkWOWOFORM >();
            if ( form->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_wowo_chunks.push_back( form );
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "WVRN",
                                                                                        "WFNT",
                                                                                        "WINF",
                                                                                        "WDOC",
                                                                                        "WSMD",
                                                                                        "WPF0",
                                                                                        "WTAB",
                                                                                        "WHED",
                                                                                        "WFOT",
                                                                                        "WPAL",
                                                                                        "WKCO",
                                                                                        "WPF1",
                                                                                        "WSTY",
                                                                                        "WSTT",
                                                                                        "WFSC",
                                                                                        "WSPC" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            LOG_DEBUG( "got generic chunk %.*s of %u bytes", 4, buf, chunk_parsed_len );

            m_wowo_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "WordWorth document" );

    return 0;
}

int ChunkFORM::parseILBM( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "ILBM" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        LOG_DEBUG( "Decode %.*s chunk of %u bytes", 4, buf, chunklen );

        if ( chunktype == CHUNKNAME( "BMHD" ) ) {
            auto chunk = std::make_shared< ChunkBMHD >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.bmhd = chunk;

            if ( chunk->fill_info( m_info ) != 0 ) {
                LOG_DEBUG( "failed to get info from BMHD header" );
            }
        } else if ( chunktype == CHUNKNAME( "CMAP" ) ) {
            auto chunk = std::make_shared< ChunkCMAP >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.cmap = chunk;
        } else if ( chunktype == CHUNKNAME( "GRAB" ) ) {
            auto chunk = std::make_shared< ChunkGRAB >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.grab = chunk;
        } else if ( chunktype == CHUNKNAME( "DEST" ) ) {
            auto chunk = std::make_shared< ChunkDEST >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.dest = chunk;
        } else if ( chunktype == CHUNKNAME( "SPRT" ) ) {
            auto chunk = std::make_shared< ChunkSPRT >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.sprt = chunk;
        } else if ( chunktype == CHUNKNAME( "CAMG" ) ) {
            auto chunk = std::make_shared< ChunkCAMG >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.camg = chunk;
        } else if ( chunktype == CHUNKNAME( "BODY" ) ) {
            auto chunk = std::make_shared< ChunkBODY >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.body = chunk;
        } else if ( chunktype == CHUNKNAME( "ANNO" ) ) {
            auto chunk = std::make_shared< ChunkANNO >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.anno = chunk;
        } else if ( chunktype == CHUNKNAME( "DPI " ) ) {
            auto chunk = std::make_shared< ChunkDPI >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.dpi = chunk;
        } else if ( chunktype == CHUNKNAME( "CRNG" ) ) {
            auto chunk = std::make_shared< ChunkCRNG >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.crng = chunk;
        } else if ( chunktype == CHUNKNAME( "ANHD" ) ) {
            auto chunk = std::make_shared< ChunkANHD >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.anhd = chunk;
        } else if ( chunktype == CHUNKNAME( "DLTA" ) ) {
            auto chunk = std::make_shared< ChunkDLTA >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ilbm_chunks.dlta = chunk;
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "CCRT",
                                                                                        "CLUT",
                                                                                        "DYCP",
                                                                                        "CTBL",
                                                                                        "DPPV",
                                                                                        "DRNG",
                                                                                        "EPSF",
                                                                                        "DPPS",
                                                                                        "MPIC" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) == 0 ) {
                LOG_DEBUG( "got known generic chunk %.*s of %u bytes", 4, buf, chunk_parsed_len );

                m_ilbm_chunks.known_chunks.push_back( chunk );
            } else {
                chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{} );
                if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                    LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                    if ( fail_on_unsupported() ) {
                        return -EINVAL;
                    }
                }

                LOG_DEBUG( "got generic chunk %.*s of %u bytes", 4, buf, chunk_parsed_len );

                m_ilbm_chunks.other_chunks.push_back( chunk );
            }
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "ILBM image" );

    return 0;
}

int ChunkFORM::parseDEEP( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "DEEP" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        LOG_DEBUG( "Decode %.*s chunk of %u bytes", 4, buf, chunklen );

        if ( chunktype == CHUNKNAME( "DGBL" ) ) {
            auto chunk = std::make_shared< ChunkDGBL >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_deep_chunks.dgbl = chunk;

            if ( chunk->fill_info( m_info ) != 0 ) {
                LOG_DEBUG( "failed to get info from DGBL header" );
            }
        } else if ( chunktype == CHUNKNAME( "DPEL" ) ) {
            auto chunk = std::make_shared< ChunkDPEL >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_deep_chunks.dpel = chunk;
        } else if ( chunktype == CHUNKNAME( "DLOC" ) ) {
            auto chunk = std::make_shared< ChunkDLOC >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_deep_chunks.dloc = chunk;
        } else if ( chunktype == CHUNKNAME( "DBOD" ) ) {
            auto chunk = std::make_shared< ChunkDBOD >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_deep_chunks.dbod = chunk;
        } else if ( chunktype == CHUNKNAME( "DCHG" ) ) {
            auto chunk = std::make_shared< ChunkDCHG >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_deep_chunks.dchg = chunk;
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "MPIC" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) == 0 ) {
                LOG_DEBUG( "got known generic chunk %.*s of %u bytes", 4, buf, chunk_parsed_len );

                m_deep_chunks.known_chunks.push_back( chunk );
            } else {
                chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{} );
                if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                    LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                    if ( fail_on_unsupported() ) {
                        return -EINVAL;
                    }
                }

                LOG_DEBUG( "got generic chunk %.*s of %u bytes", 4, buf, chunk_parsed_len );

                m_deep_chunks.other_chunks.push_back( chunk );
            }
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "DEEP image" );

    return 0;
}

int ChunkFORM::parseSWRT( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "SWRT" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "CHRS" ) ) {
            LOG_DEBUG( "Got CHRS chunk of %u bytes", chunklen );
            auto chrs = std::make_shared< ChunkCHRS >();
            if ( chrs->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_swrt_chunks.push_back( chrs );
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "RULE",
                                                                                        "ATTR",
                                                                                        "LMST",
                                                                                        "GINF",
                                                                                        "FDTA",
                                                                                        "SHD1",
                                                                                        "RMST",
                                                                                        "HYPH",
                                                                                        "ASCI",
                                                                                        "SPEL",
                                                                                        "PRGH",
                                                                                        "DOC ",
                                                                                        "PICP",
                                                                                        "LINP",
                                                                                        "BOXP",
                                                                                        "OVLP",
                                                                                        "SWCL",
                                                                                        "IDXH",
                                                                                        "IDXD",
                                                                                        "TOCH",
                                                                                        "TOCD",
                                                                                        "IDXP",
                                                                                        "SEC1",
                                                                                        "PAG3",
                                                                                        "PRN3",
                                                                                        "TBLK",
                                                                                        "OUTH",
                                                                                        "OUTD",
                                                                                        "FNTH",
                                                                                        "FNTD",
                                                                                        "BIBH",
                                                                                        "BIBD",
                                                                                        "TOIH",
                                                                                        "TOID",
                                                                                        "TSSH",
                                                                                        "TSSD",
                                                                                        "PASH",
                                                                                        "PASD",
                                                                                        "SAVP",
                                                                                        "TOCP",
                                                                                        "OUTP",
                                                                                        "BIBP",
                                                                                        "TOIP",
                                                                                        "ENDP",
                                                                                        "DSP2",
                                                                                        "HFDA",
                                                                                        "PNDA",
                                                                                        "SHPP",
                                                                                        "DINF",
                                                                                        "TBDY",
                                                                                        "TABS"} );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_swrt_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    if ( ! attribute( "chrs.newline" ).second ) {
        set_attribute( "chrs.newline", 1 );
    }

    m_info.set_description( "FinalWriter document" );

    return 0;
}

int ChunkFORM::parseFANT( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "FANT" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "CSTR" ) ) {
            LOG_DEBUG( "Got CSTR chunk of %u bytes", chunklen );
            auto chrs = std::make_shared< ChunkCSTR >();
            if ( chrs->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_fant_chunks.chunks.push_back( chrs );
        } else if ( chunktype == CHUNKNAME( "FORM" ) ) {
            LOG_DEBUG( "Got FORM chunk of %u bytes", chunklen );
            auto form = std::make_shared< ChunkFORM >();
            if ( form->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_fant_chunks.chunks.push_back( form );
        } else if ( chunktype == CHUNKNAME( "FHDR" ) ) {
            LOG_DEBUG( "Got FDHR chunk of %u bytes", chunklen );
            auto chunk = std::make_shared< ChunkFHDR >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_fant_chunks.fhdr = chunk;

            if ( chunk->fill_info( m_info ) != 0 ) {
                LOG_DEBUG( "failed to get info from FHDR header" );
            }
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "FRAM",
                                                                                        "POLY",
                                                                                        "SEFX" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_fant_chunks.chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "FantaVision animation" );

    return 0;
}

int ChunkFORM::parse8SVX( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "8SVX" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "ANNO" ) ) {
            LOG_DEBUG( "Got ANNO chunk of %u bytes", chunklen );
            auto anno = std::make_shared< ChunkANNO >();
            if ( anno->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_8svx_chunks.anno = anno;
        } else if ( chunktype == CHUNKNAME( "NAME" ) ) {
            LOG_DEBUG( "Got NAME chunk of %u bytes", chunklen );
            auto name = std::make_shared< ChunkNAME >();
            if ( name->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_8svx_chunks.name = name;
        } else if ( chunktype == CHUNKNAME( "VHDR" ) ) {
            LOG_DEBUG( "Got VHDR chunk of %u bytes", chunklen );
            auto vhdr = std::make_shared< ChunkVHDR >();
            if ( vhdr->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_8svx_chunks.vhdr = vhdr;

            if ( vhdr->fill_info( m_info ) != 0 ) {
                LOG_DEBUG( "failed to get info from BMHD header" );
            }
        } else if ( chunktype == CHUNKNAME( "BODY" ) ) {
            LOG_DEBUG( "Got BODY chunk of %u bytes", chunklen );
            auto body = std::make_shared< Chunk8SVXBODY >();
            if ( body->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_8svx_chunks.body = body;
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "BODY",
                                                                                        "ATAK",
                                                                                        "RLSE",
                                                                                        "AUTH",
                                                                                        "(C) ",
                                                                                        "CHAN",
                                                                                        "SEFX",
                                                                                        "SEVH" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_8svx_chunks.other_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    if ( ! attribute( "8svx.ignore_volume" ).second ) {
        set_attribute( "8svx.ignore_volume", 0 );
    }

    m_info.set_description( "8SVX sound sample file" );

    return 0;
}

int ChunkFORM::parseANIM( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "ANIM" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "FORM" ) ) {
            LOG_DEBUG( "Got FORM chunk of %u bytes", chunklen );
            auto form = std::make_shared< ChunkFORM >();
            if ( form->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_anim_chunks.form_chunks.push_back( form );
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_anim_chunks.other_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    if ( ! attribute( "anim.duplicate_frames" ).second ) {
        set_attribute( "anim.duplicate_frames", 0 );
    }

    m_info.set_description( "ANIM animation" );

    return 0;
}

int ChunkFORM::parsePREF( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "PREF" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "PREF" ) ) {
            LOG_DEBUG( "Got PREF chunk of %u bytes", chunklen );
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "PRHD" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_pref_chunks.push_back( chunk );
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{} );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_pref_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "Amiga preferences file" );

    return 0;
}

int ChunkFORM::parseCTLG( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "CTLG" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "FVER" ) ) {
            LOG_DEBUG( "Got FVER chunk of %u bytes", chunklen );
            auto chunk = std::make_shared< ChunkFVER >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ctlg_chunks.push_back( chunk );
        } else if ( chunktype == CHUNKNAME( "LANG" ) ) {
            LOG_DEBUG( "Got LANG chunk of %u bytes", chunklen );
            auto chunk = std::make_shared< ChunkLANG >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ctlg_chunks.push_back( chunk );
        } else if ( chunktype == CHUNKNAME( "ANNO" ) ) {
            LOG_DEBUG( "Got ANNO chunk of %u bytes", chunklen );
            auto chunk = std::make_shared< ChunkANNO >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_ctlg_chunks.push_back( chunk );
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "CSET",
                                                                                        "STRS" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_ctlg_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "Amiga language translation file" );

    return 0;
}

int ChunkFORM::parseCDVR( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "CDVR" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "ANNO" ) ) {
            LOG_DEBUG( "Got ANNO chunk of %u bytes", chunklen );
            auto chunk = std::make_shared< ChunkANNO >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_cdvr_chunks.push_back( chunk );
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "VARS" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_cdvr_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    return 0;
}

int ChunkFORM::parseREFL( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "REFL" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        //auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "INFO",
                                                                                    "DRE2",
                                                                                    "GMAT",
                                                                                    "SURF",
                                                                                    "TXM1",
                                                                                    "RKA2",
                                                                                    "POLY",
                                                                                    "TXO1",
                                                                                    "LGH3",
                                                                                    "NEBL",
                                                                                    "SEQ1",
                                                                                    "POL2",
                                                                                    "RBEZ" } );
        if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
            LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
            if ( fail_on_unsupported() ) {
                return -EINVAL;
            }
        }

        m_refl_chunks.push_back( chunk );

        // Reflections 3 does not align chunks at short words!
        buf += 8 + chunklen;
        buflen -= 8 + chunklen;
    }

    m_info.set_description( "Reflections 3D scene file" );

    return 0;
}

int ChunkFORM::parseGXUI( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "GXUI" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "FORM" ) ) {
            LOG_DEBUG( "Got FORM chunk of %u bytes", chunklen );
            auto chunk = std::make_shared< ChunkGXUIFORM >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_gxui_chunks.push_back( chunk );
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "GGUI" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_gxui_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "GadtoolsBox UI file" );

    return 0;
}

int ChunkFORM::parseSMUS( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "SMUS" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "NAME" ) ) {
            LOG_DEBUG( "Got NAME chunk of %u bytes", chunklen );
            auto chunk = std::make_shared< ChunkNAME >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_smus_chunks.push_back( chunk );
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "SHDR",
                                                                                        "SNX1",
                                                                                        "TRAK",
                                                                                        "INS1" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_smus_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "Simple Music file" );

    return 0;
}

int ChunkFORM::parseGXWD( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "GXWD" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "FORM" ) ) {
            LOG_DEBUG( "Got FORM chunk of %u bytes", chunklen );
            auto chunk = std::make_shared< ChunkFORM >();
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_generic_chunks.push_back( chunk );
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "WDDA",
                                                                                        "FORM" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            m_generic_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "GadtoolsBox Window file" );

    return 0;
}

int ChunkFORM::parseKICK( const char *buf,
                          size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    auto type = betoh_u32( get_u32( buf, 0 ) );

    if ( type != CHUNKNAME( "KICK" ) ) {
        return -EINVAL;
    }

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunktype == CHUNKNAME( "CEND" ) ) {
            LOG_DEBUG( "found chunk end" );
            break;
        }

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length %u", chunklen );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "KSHD",
                                                                                    "ASUM",
                                                                                    "PTAB",
                                                                                    "PAT2",
                                                                                    "CPAT",
                                                                                    "BREL",
                                                                                    "CEND" } );
        if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
            LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
            if ( fail_on_unsupported() ) {
                return -EINVAL;
            }
        }

        m_generic_chunks.push_back( chunk );

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    m_info.set_description( "Kickstart file" );

    return 0;
}

int ChunkFORM::parseGeneric( const char *buf,
                             size_t buflen )
{
    if ( buflen < 4 ) {
        return -EINVAL;
    }

    //auto type = betoh_u32( get_u32( buf, 0 ) );

    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        //auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        LOG_DEBUG( "Decode %.*s chunk of %u bytes", 4, buf, chunklen );

        auto chunk = check_and_parse( buf,
                                      buflen,
                                      chunk_parsed_len );
        if ( ! chunk ) {
            LOG_CRITICAL( "failed to get chunk" );
            return -EINVAL;
        }

        LOG_DEBUG( "got generic chunk %.*s of %u bytes", 4, buf, chunk_parsed_len );

        m_generic_chunks.push_back( chunk );

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    return 0;
}

int ChunkFORM::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();
    const char *buf = &data[0];

    if ( chunklen < 12 ) {
        LOG_DEBUG( "too short for FORM" );
        return -EINVAL;
    }

    if ( m_chunk_type != CHUNKNAME( "FORM" ) ) {
        LOG_DEBUG( "not a FORM chunk" );
        return -EINVAL;
    }

    LOG_DEBUG( "Found FORM chunk" );
    LOG_DEBUG( "parse %u bytes...", chunklen );

    m_form_type = betoh_u32( get_u32( buf, 0 ) );
    auto form_string = std::string( buf, 4 );

    if ( m_form_type == CHUNKNAME( "FTXT" ) ) {
        if ( parseFTXT( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "WOWO" ) ) {
        if ( parseWOWO( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "ILBM" ) ) {
        if ( parseILBM( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "DEEP" ) ) {
        if ( parseDEEP( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "SWRT" ) ) {
        if ( parseSWRT( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "FANT" ) ) {
        if ( parseFANT( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "8SVX" ) ) {
        if ( parse8SVX( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "ANIM" ) ) {
        if ( parseANIM( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "RGB8" ) ||
                m_form_type == CHUNKNAME( "RGBN" ) ||
                m_form_type == CHUNKNAME( "ACBM" ) ||
                m_form_type == CHUNKNAME( "DR2D" ) ||
                m_form_type == CHUNKNAME( "FNTR" ) ||
                m_form_type == CHUNKNAME( "FNTV" ) ||
                m_form_type == CHUNKNAME( "PICS" ) ||
                m_form_type == CHUNKNAME( "TDDD" ) ||
                m_form_type == CHUNKNAME( "YUVN" ) ) {
        // other picture forms
        if ( parseGeneric( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "ANBM" ) ) {
        // other anim forms
        if ( parseGeneric( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "VDEO" ) ) {
        // other video forms
        if ( parseGeneric( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "AIFF" ) ||
                m_form_type == CHUNKNAME( "SAMP" ) ||
                m_form_type == CHUNKNAME( "UVOX" ) ) {
        // other sound forms
        if ( parseGeneric( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "GSCR" ) ||
                m_form_type == CHUNKNAME( "TRAK" ) ||
                m_form_type == CHUNKNAME( "USCR" ) ) {
        // other music forms
        if ( parseGeneric( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "WORD" ) ||
                m_form_type == CHUNKNAME( "(C) " ) ||
                m_form_type == CHUNKNAME( "AUTH" ) ||
                m_form_type == CHUNKNAME( "DOC " ) ||
                m_form_type == CHUNKNAME( "FOOT" ) ||
                m_form_type == CHUNKNAME( "HEAD" ) ||
                m_form_type == CHUNKNAME( "PAGE" ) ||
                m_form_type == CHUNKNAME( "PARA" ) ||
                m_form_type == CHUNKNAME( "PDEF" ) ||
                m_form_type == CHUNKNAME( "TABS" ) ||
                m_form_type == CHUNKNAME( "TEXT" ) ||
                m_form_type == CHUNKNAME( "VERS" ) ) {
        // other text forms
        if ( parseGeneric( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "PREF" ) ) {
        // preferences
        if ( parsePREF( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "CTLG" ) ) {
        // catalog files
        if ( parseCTLG( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "CDVR" ) ) {
        // cando variables files
        if ( parseCDVR( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "REFL" ) ) {
        // Reflection 3 files
        if ( parseREFL( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "GXUI" ) ) {
        // gadget toolbox
        if ( parseGXUI( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "SMUS" ) ) {
        // Simple Music files
        if ( parseSMUS( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "GXWD" ) ) {
        // gadtools
        if ( parseGXWD( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( m_form_type == CHUNKNAME( "KICK" ) ) {
        // 
        if ( parseKICK( buf, chunklen ) != 0 ) {
            return -EINVAL;
        }
    } else if ( parseKnownTypes( m_form_type, buf, chunklen ) == 0 ) {
    } else {
        LOG_DEBUG( "Unsupported type %s, skipping", form_string.c_str() );
        if ( fail_on_unsupported() ) {
            return -EINVAL;
        } else {
            if ( parseGeneric( buf, chunklen ) != 0 ) {
                return -EINVAL;
            }
        }
    }

    return 0;
}

std::string ChunkFORM::dump_as_readable_text( const shared_properties_t &shared_props )
{
    std::string res;

    switch ( m_form_type ) {
        case CHUNKNAME( "FTXT" ): {
            for ( auto &ftxt_chunk : m_ftxt_chunks ) {
                res += ftxt_chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "WOWO" ): {
            for ( auto &wowo_chunk : m_wowo_chunks ) {
                res += wowo_chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "ILBM" ): {
            if ( m_ilbm_chunks.anno ) {
                res += m_ilbm_chunks.anno->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "SWRT" ): {
            for ( auto &swrt_chunk : m_swrt_chunks ) {
                res += swrt_chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "FANT" ): {
            if ( m_fant_chunks.fhdr ) {
                res += m_fant_chunks.fhdr->dump_as_readable_text( shared_props );
            }
            for ( auto &fant_chunk : m_fant_chunks.chunks ) {
                res += fant_chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "8SVX" ): {
            if ( m_8svx_chunks.vhdr ) {
                res += m_8svx_chunks.vhdr->dump_as_readable_text( shared_props );
            }
            if ( m_8svx_chunks.name ) {
                res += m_8svx_chunks.name->dump_as_readable_text( shared_props );
            }
            if ( m_8svx_chunks.anno ) {
                res += m_8svx_chunks.anno->dump_as_readable_text( shared_props );
            }
            if ( m_8svx_chunks.body ) {
                res += m_8svx_chunks.body->dump_as_readable_text( shared_props );
            }
            for ( auto &svx_chunk : m_8svx_chunks.other_chunks ) {
                res += svx_chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "ANIM" ): {
            for ( auto &form_chunk : m_anim_chunks.form_chunks ) {
                res += form_chunk->dump_as_readable_text( shared_props );
            }
            for ( auto &chunk : m_anim_chunks.other_chunks ) {
                res += chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "PREF" ): {
            for ( auto &pref_chunk : m_pref_chunks ) {
                res += pref_chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "CTLG" ): {
            for ( auto &ctlg_chunk : m_ctlg_chunks ) {
                res += ctlg_chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "CDVR" ): {
            for ( auto &cdvr_chunk : m_cdvr_chunks ) {
                res += cdvr_chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "REFL" ): {
            for ( auto &chunk : m_refl_chunks ) {
                res += chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "GXUI" ): {
            for ( auto &chunk : m_gxui_chunks ) {
                res += chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "SMUS" ): {
            for ( auto &chunk : m_smus_chunks ) {
                res += chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
        case CHUNKNAME( "GXWD" ):
        case CHUNKNAME( "KICK" ):
        default: {
            for ( auto &chunk : m_generic_chunks ) {
                res += chunk->dump_as_readable_text( shared_props );
            }
            break;
        }
    }

    return res;
}

json_t *ChunkFORM::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG("called");

    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "FORM" ) );

    json_object_set_new( jroot, "subtype", json_string( CHUNKTYPE2NAME( m_form_type ).c_str() ) );

    auto jarray = json_array();

    switch ( m_form_type ) {
        case CHUNKNAME( "FTXT" ): {
            for ( auto &ftxt_chunk : m_ftxt_chunks ) {
                auto jentry = ftxt_chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "WOWO" ): {
            for ( auto &wowo_chunk : m_wowo_chunks ) {
                auto jentry = wowo_chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "ILBM" ): {
            const auto bmhd = m_ilbm_chunks.bmhd ? m_ilbm_chunks.bmhd : get_shared_chunk< ChunkBMHD >( "ILBM", "BMHD", shared_props );
            if ( bmhd ) {
                auto jentry = bmhd->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto cmap = m_ilbm_chunks.cmap ? m_ilbm_chunks.cmap : get_shared_chunk< ChunkCMAP >( "ILBM", "CMAP", shared_props );
            if ( cmap ) {
                auto jentry = cmap->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto grab = m_ilbm_chunks.grab ? m_ilbm_chunks.grab : get_shared_chunk< ChunkGRAB >( "ILBM", "GRAB", shared_props );
            if ( grab ) {
                auto jentry = grab->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto dest = m_ilbm_chunks.dest ? m_ilbm_chunks.dest : get_shared_chunk< ChunkDEST >( "ILBM", "DEST", shared_props );
            if ( dest ) {
                auto jentry = dest->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto sprt = m_ilbm_chunks.sprt ? m_ilbm_chunks.sprt : get_shared_chunk< ChunkSPRT >( "ILBM", "SPRT", shared_props );
            if ( sprt ) {
                auto jentry = sprt->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto camg = m_ilbm_chunks.camg ? m_ilbm_chunks.camg : get_shared_chunk< ChunkCAMG >( "ILBM", "CAMG", shared_props );
            if ( camg ) {
                auto jentry = camg->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto body = m_ilbm_chunks.body ? m_ilbm_chunks.body : get_shared_chunk< ChunkBODY >( "ILBM", "BODY", shared_props );
            if ( body ) {
                auto jentry = body->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto anno = m_ilbm_chunks.anno ? m_ilbm_chunks.anno : get_shared_chunk< ChunkANNO >( "ILBM", "ANNO", shared_props );
            if ( anno ) {
                auto jentry = anno->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto dpi = m_ilbm_chunks.dpi ? m_ilbm_chunks.dpi : get_shared_chunk< ChunkDPI >( "ILBM", "DPI ", shared_props );
            if ( dpi ) {
                auto jentry = dpi->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto crng = m_ilbm_chunks.crng ? m_ilbm_chunks.crng : get_shared_chunk< ChunkCRNG >( "ILBM", "CRNG", shared_props );
            if ( crng ) {
                auto jentry = crng->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto anhd = m_ilbm_chunks.anhd ? m_ilbm_chunks.anhd : get_shared_chunk< ChunkANHD >( "ILBM", "ANHD", shared_props );
            if ( anhd ) {
                auto jentry = anhd->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto dlta = m_ilbm_chunks.dlta ? m_ilbm_chunks.dlta : get_shared_chunk< ChunkDLTA >( "ILBM", "DLTA", shared_props );
            if ( dlta ) {
                auto jentry = dlta->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            for ( auto &chunk : m_ilbm_chunks.known_chunks ) {
                auto jentry = chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            for ( auto &chunk : m_ilbm_chunks.other_chunks ) {
                auto jentry = chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            break;
        }
        case CHUNKNAME( "DEEP" ): {
            const auto dgbl = m_deep_chunks.dgbl ? m_deep_chunks.dgbl : get_shared_chunk< ChunkDGBL >( "DEEP", "DGBL", shared_props );
            if ( dgbl ) {
                auto jentry = dgbl->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto dpel = m_deep_chunks.dpel ? m_deep_chunks.dpel : get_shared_chunk< ChunkDPEL >( "DEEP", "DPEL", shared_props );
            if ( dpel ) {
                auto jentry = dpel->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto dloc = m_deep_chunks.dloc ? m_deep_chunks.dloc : get_shared_chunk< ChunkDLOC >( "DEEP", "DLOC", shared_props );
            if ( dloc ) {
                auto jentry = dloc->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto dbod = m_deep_chunks.dbod ? m_deep_chunks.dbod : get_shared_chunk< ChunkDBOD >( "DEEP", "DBOD", shared_props );
            if ( dbod ) {
                auto jentry = dbod->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            const auto dchg = m_deep_chunks.dchg ? m_deep_chunks.dchg : get_shared_chunk< ChunkDCHG >( "DEEP", "DCHG", shared_props );
            if ( dchg ) {
                auto jentry = dchg->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            for ( auto &chunk : m_deep_chunks.known_chunks ) {
                auto jentry = chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            for ( auto &chunk : m_deep_chunks.other_chunks ) {
                auto jentry = chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }

            break;
        }
        case CHUNKNAME( "SWRT" ): {
            for ( auto &swrt_chunk : m_swrt_chunks ) {
                auto jentry = swrt_chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "FANT" ): {
            if ( m_fant_chunks.fhdr ) {
                auto jentry = m_fant_chunks.fhdr->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            for ( auto &fant_chunk : m_fant_chunks.chunks ) {
                auto jentry = fant_chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "8SVX" ): {
            if ( m_8svx_chunks.vhdr ) {
                auto jentry = m_8svx_chunks.vhdr->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            if ( m_8svx_chunks.name ) {
                auto jentry = m_8svx_chunks.name->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            if ( m_8svx_chunks.anno ) {
                auto jentry = m_8svx_chunks.anno->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            if ( m_8svx_chunks.body ) {
                auto jentry = m_8svx_chunks.body->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            for ( auto &svx_chunk : m_8svx_chunks.other_chunks ) {
                auto jentry = svx_chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "ANIM" ): {
            for ( auto &form_chunk : m_anim_chunks.form_chunks ) {
                auto jentry = form_chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            for ( auto &chunk : m_anim_chunks.other_chunks ) {
                auto jentry = chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "PREF" ): {
            for ( auto &pref_chunk : m_pref_chunks ) {
                auto jentry = pref_chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "CTLG" ): {
            for ( auto &ctlg_chunk : m_ctlg_chunks ) {
                auto jentry = ctlg_chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "CDVR" ): {
            for ( auto &cdvr_chunk : m_cdvr_chunks ) {
                auto jentry = cdvr_chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "REFL" ): {
            for ( auto &chunk : m_refl_chunks ) {
                auto jentry = chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "GXUI" ): {
            for ( auto &chunk : m_gxui_chunks ) {
                auto jentry = chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "SMUS" ): {
            for ( auto &chunk : m_smus_chunks ) {
                auto jentry = chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
        }
        case CHUNKNAME( "GXWD" ):
        case CHUNKNAME( "KICK" ):
        default:
            for ( auto &chunk : m_generic_chunks ) {
                auto jentry = chunk->dump_as_json( shared_props, include_raw );

                json_array_append_new( jarray, jentry );
            }
            break;
    }

    json_object_set_new( jroot, "content", jarray );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

int ChunkFORM::dump_as_binary( const shared_properties_t &shared_props,
                               std::list< IFFBinaryResult > &output,
                               bool include_raw ) const
{
    switch ( m_form_type ) {
        case CHUNKNAME( "FTXT" ): {
            std::string parsed_text;
            std::vector< char > raw_data;
            for ( auto &ftxt_chunk : m_ftxt_chunks ) {
                parsed_text += ftxt_chunk->dump_as_readable_text( shared_props );

                if ( include_raw ) {
                    std::vector< char > tdata;
                    if ( ftxt_chunk->dump_as_raw( shared_props, tdata ) == 0 ) {
                        for ( auto ch : tdata ) {
                            raw_data.push_back( ch );
                        }
                    } else {
                        LOG_CRITICAL( "failed to dump raw data" );
                    }
                }
            }

            IFFBinaryResult res1;
            IFFBinaryResult res2;

            res1.file_ending = "txt";
            for ( auto ch : parsed_text ) {
                res1.data.push_back( ch );
            }

            output.push_back( res1 );

            if ( include_raw ) {
                res2.file_ending = "raw";
                res2.data = raw_data;

                output.push_back( res2 );
            }

            for ( auto &chunk : m_ftxt_chunks ) {
                if ( chunk->dump_as_binary( shared_props,
                                            output,
                                            include_raw ) != 0 ) {
                    LOG_DEBUG( "failed to dump binary data" );
                }
            }

            break;
        }
        case CHUNKNAME( "WOWO" ): {
            std::string parsed_text;
            std::vector< char > raw_data;
            for ( auto &wowo_chunk : m_wowo_chunks ) {
                parsed_text += wowo_chunk->dump_as_readable_text( shared_props );

                if ( include_raw ) {
                    std::vector< char > tdata;
                    if ( wowo_chunk->dump_as_raw( shared_props, tdata ) == 0 ) {
                        for ( auto ch : tdata ) {
                            raw_data.push_back( ch );
                        }
                    } else {
                        LOG_CRITICAL( "failed to dump raw data" );
                    }
                }
            }

            IFFBinaryResult res1;
            IFFBinaryResult res2;

            res1.file_ending = "txt";
            for ( auto ch : parsed_text ) {
                res1.data.push_back( ch );
            }

            output.push_back( res1 );

            if ( include_raw ) {
                res2.file_ending = "raw";
                res2.data = raw_data;

                output.push_back( res2 );
            }

            for ( auto &chunk : m_wowo_chunks ) {
                if ( chunk->dump_as_binary( shared_props,
                                            output,
                                            include_raw ) != 0 ) {
                    LOG_DEBUG( "failed to dump binary data" );
                }
            }

            break;
        }
        case CHUNKNAME( "ILBM" ): {
            return dump_ilbm_as_binary( shared_props, output, include_raw );
        }
        case CHUNKNAME( "DEEP" ): {
            return dump_deep_as_binary( shared_props, output, include_raw );
        }
        case CHUNKNAME( "SWRT" ): {
            std::string parsed_text;
            std::vector< char > raw_data;
            for ( auto &swrt_chunk : m_swrt_chunks ) {
                parsed_text += swrt_chunk->dump_as_readable_text( shared_props );

                if ( include_raw ) {
                    std::vector< char > tdata;
                    if ( swrt_chunk->dump_as_raw( shared_props, tdata ) == 0 ) {
                        for ( auto ch : tdata ) {
                            raw_data.push_back( ch );
                        }
                    } else {
                        LOG_CRITICAL( "failed to dump raw data" );
                    }
                }
            }

            IFFBinaryResult res1;
            IFFBinaryResult res2;

            res1.file_ending = "txt";
            for ( auto ch : parsed_text ) {
                res1.data.push_back( ch );
            }

            output.push_back( res1 );

            if ( include_raw ) {
                res2.file_ending = "raw";
                res2.data = raw_data;

                output.push_back( res2 );
            }

            for ( auto &chunk : m_swrt_chunks ) {
                if ( chunk->dump_as_binary( shared_props,
                                            output,
                                            include_raw ) != 0 ) {
                    LOG_DEBUG( "failed to dump binary data" );
                }
            }
            break;
        }
        case CHUNKNAME( "FANT" ): {
            std::shared_ptr< ChunkCMAP > cmap = nullptr;

            std::string parsed_text;
            std::vector< char > raw_data;

            if ( m_fant_chunks.fhdr ) {
                parsed_text += m_fant_chunks.fhdr->dump_as_readable_text( shared_props );

                cmap = m_fant_chunks.fhdr->get_cmap();
            }

            shared_properties_t my_shared_props = shared_props;
            if ( cmap ) {
                my_shared_props[ std::make_pair( "ILBM", "CMAP" ) ] = cmap;
            }

            for ( auto &fant_chunk : m_fant_chunks.chunks ) {
                parsed_text += fant_chunk->dump_as_readable_text( my_shared_props );

                if ( include_raw ) {
                    std::vector< char > tdata;
                    if ( fant_chunk->dump_as_raw( my_shared_props, tdata ) == 0 ) {
                        for ( auto ch : tdata ) {
                            raw_data.push_back( ch );
                        }
                    } else {
                        LOG_CRITICAL( "failed to dump raw data" );
                    }
                }
            }

            IFFBinaryResult res1;
            IFFBinaryResult res2;

            res1.file_ending = "txt";
            for ( auto ch : parsed_text ) {
                res1.data.push_back( ch );
            }

            output.push_back( res1 );

            if ( include_raw ) {
                res2.file_ending = "raw";
                res2.data = raw_data;

                output.push_back( res2 );
            }

            for ( auto &chunk : m_fant_chunks.chunks ) {
                if ( chunk->dump_as_binary( my_shared_props,
                                            output,
                                            include_raw ) != 0 ) {
                    LOG_DEBUG( "failed to dump binary data" );
                }
            }

            break;
        }
        case CHUNKNAME( "8SVX" ): {
            return dump_8svx_as_binary( shared_props, output, include_raw );
        }
        case CHUNKNAME( "ANIM" ): {
            return dump_anim_as_binary( shared_props, output, include_raw );
        }
        case CHUNKNAME( "CTLG" ): {
            for ( auto &chunk : m_ctlg_chunks ) {
                if ( chunk->dump_as_binary( shared_props,
                                            output,
                                            include_raw ) != 0 ) {
                    LOG_DEBUG( "failed to dump binary data" );
                }
            }
            break;
        }
        case CHUNKNAME( "CDVR" ): {
            for ( auto &chunk : m_cdvr_chunks ) {
                if ( chunk->dump_as_binary( shared_props,
                                            output,
                                            include_raw ) != 0 ) {
                    LOG_DEBUG( "failed to dump binary data" );
                }
            }
            break;
        }
        case CHUNKNAME( "REFL" ): {
            for ( auto &chunk : m_refl_chunks ) {
                if ( chunk->dump_as_binary( shared_props,
                                            output,
                                            include_raw ) != 0 ) {
                    LOG_DEBUG( "failed to dump binary data" );
                }
            }
            break;
        }
        case CHUNKNAME( "GXUI" ): {
            for ( auto &chunk : m_gxui_chunks ) {
                if ( chunk->dump_as_binary( shared_props,
                                            output,
                                            include_raw ) != 0 ) {
                    LOG_DEBUG( "failed to dump binary data" );
                }
            }
            break;
        }
        case CHUNKNAME( "SMUS" ): {
            for ( auto &chunk : m_smus_chunks ) {
                if ( chunk->dump_as_binary( shared_props,
                                            output,
                                            include_raw ) != 0 ) {
                    LOG_DEBUG( "failed to dump binary data" );
                }
            }
            break;
        }
        case CHUNKNAME( "GXWD" ):
        case CHUNKNAME( "KICK" ):
        default: {
            for ( auto &chunk : m_generic_chunks ) {
                if ( chunk->dump_as_binary( shared_props,
                                            output,
                                            include_raw ) != 0 ) {
                    LOG_DEBUG( "failed to dump binary data" );
                }
            }
            break;
        }
    }

    return 0;
}

int ChunkFORM::convert_ilbm_to_pixels( const shared_properties_t &shared_props,
                                       std::vector< uint32_t > &pixels ) const
{
    const auto bmhd = m_ilbm_chunks.bmhd ? m_ilbm_chunks.bmhd : get_shared_chunk< ChunkBMHD >( "ILBM", "BMHD", shared_props );
    const auto body = m_ilbm_chunks.body ? m_ilbm_chunks.body : get_shared_chunk< ChunkBODY >( "ILBM", "BODY", shared_props );

    if ( ! bmhd ) {
        LOG_CRITICAL( "no ILBM BMDH chunk" );
        return -EINVAL;
    }
    if ( ! body ) {
        LOG_CRITICAL( "no ILBM BODY chunk" );
        return -EINVAL;
    }

    int res = body->decode( bmhd, pixels );

    if ( res != 0 ) {
        LOG_CRITICAL( "error decoding body: %d", res );
        return res;
    }

    return 0;
}

int ChunkFORM::convert_pixels_to_rgba( const std::shared_ptr< ChunkBMHD > &bmhd,
                                       const std::shared_ptr< ChunkCMAP > &cmap,
                                       const std::shared_ptr< ChunkCAMG > &camg,
                                       const std::vector< uint32_t > &pixels,
                                       std::vector< uint32_t > &rgba_pixels )
{
    if ( ! bmhd ) {
        LOG_CRITICAL( "no ILBM BMDH chunk" );
        return -EINVAL;
    }

    LOG_DEBUG( "got %u pixel data", pixels.size() );

    size_t pixels_written = 0;
    int ham_hold_red = 0;
    int ham_hold_green = 0;
    int ham_hold_blue = 0;

    bool ham = camg ? ( camg->display_mode() & ChunkCAMG::HAM ) : false;
    bool ehb = camg ? ( camg->display_mode() & ChunkCAMG::EXTRA_HALFBRITE ) : false;
    bool ham8 = ham && bmhd->nplanes() == 8;

    if ( ! attribute( "ilbm.adjustcolors" ).second ) {
        set_attribute( "ilbm.adjustcolors", 1 );
    }

    if ( ! attribute( "ilbm.ham6keeplowbits" ).second ) {
        set_attribute( "ilbm.ham6keeplowbits", 0 );
    }

    bool adjust_colors = false;

    size_t pixels_per_line = bmhd->line_pitch();

    if ( cmap ) {
        bool lower_bits_set = false;
        for ( size_t c = 0; c < cmap->colors(); c++ ) {
            auto col = cmap->get_color( c );

            if ( ( col.red & 0xf ) ||
                 ( col.green & 0xf ) ||
                 ( col.blue & 0xf ) ) {
                lower_bits_set = true;
                break;
            }
        }

        if ( ! lower_bits_set ) {
            LOG_VERBOSE( "color map entries seem to only have upper 4 bit values" );
            if ( attribute( "ilbm.adjustcolors" ).first ) {
                LOG_VERBOSE( "adjust colors" );
                adjust_colors = true;
            }
        }
    }

    if ( attribute( "ilbm.adjustcolors" ).first == 2 ) {
        LOG_VERBOSE( "force adjusting colors" );
        adjust_colors = true;
    }

    LOG_VERBOSE( "Decoding attributes in use:" );
    LOG_VERBOSE( "  adjustcolors: %d", attribute( "ilbm.adjustcolors" ).first );
    LOG_VERBOSE( "  ham6keeplowbits: %d", attribute( "ilbm.ham6keeplowbits" ).first );

    for ( uint32_t p : pixels ) {
        if ( bmhd->nplanes() <= 8 ) {
            if ( cmap ) {
                int red = ham_hold_red;
                int green = ham_hold_green;
                int blue = ham_hold_blue;

                if ( ham8 ) {
                    LOG_DEBUG( "decode HAM8" );
                    uint8_t mod = ( p & 0xff ) >> 6;
                    uint8_t comp = ( p & 0x3f );

                    if ( mod == 1 ) {
                        blue = ( blue & 0x03 ) | ( comp << 2 );
                    } else if ( mod == 2 ) {
                        red = ( red & 0x03 ) | ( comp << 2 );
                    } else if ( mod == 3 ) {
                        green = ( green & 0x03 ) | ( comp << 2 );
                    } else {
                        auto col = cmap->get_color( comp );

                        red = col.red;
                        green = col.green;
                        blue = col.blue;
                    }
                } else if ( ham ) {
                    LOG_DEBUG( "decode HAM" );
                    uint8_t mod = ( p & 0xff ) >> 4;
                    uint8_t comp = ( p & 0x0f );

                    if ( mod == 1 ) {
                        if ( attribute( "ilbm.ham6keeplowbits" ).first ) {
                            blue = ( blue & 0x0f ) | ( comp << 4 );
                        } else {
                            blue = ( comp << 4 ) | comp;
                        }
                    } else if ( mod == 2 ) {
                        if ( attribute( "ilbm.ham6keeplowbits" ).first ) {
                            red = ( red & 0x0f ) | ( comp << 4 );
                        } else {
                            red = ( comp << 4 ) | comp;
                        }
                    } else if ( mod == 3 ) {
                        if ( attribute( "ilbm.ham6keeplowbits" ).first ) {
                            green = ( green & 0x0f ) | ( comp << 4 );
                        } else {
                            green = ( comp << 4 ) | comp;
                        }
                    } else {
                        auto col = cmap->get_color( comp );

                        red = col.red;
                        green = col.green;
                        blue = col.blue;

                        if ( adjust_colors ) {
                            if ( ! ( bmhd->cmap_flags() & ChunkBMHD::CMAP_FLAG_8BIT ) ) {
                                red = ( red & 0xf0 ) | ( red >> 4 );
                                green = ( green & 0xf0 ) | ( green >> 4 );
                                blue = ( blue & 0xf0 ) | ( blue >> 4 );
                            }
                        }
                    }
                } else {
                    if ( ehb ) {
                        auto col = cmap->get_color( p & 0x1f );

                        if ( adjust_colors ) {
                            if ( ! ( bmhd->cmap_flags() & ChunkBMHD::CMAP_FLAG_8BIT ) ) {
                                col.red = ( col.red & 0xf0 ) | ( col.red >> 4 );
                                col.green = ( col.green & 0xf0 ) | ( col.green >> 4 );
                                col.blue = ( col.blue & 0xf0 ) | ( col.blue >> 4 );
                            }
                        }

                        if ( p & 0x20 ) {
                            red = ( col.red >> 4 ) >> 1;
                            red = ( red << 4 ) | red;
                            green = ( col.green >> 4 ) >> 1;
                            green = ( green << 4 ) | green;
                            blue = ( col.blue >> 4 ) >> 1;
                            blue = ( blue << 4 ) | blue;

                            if ( p < cmap->colors() ) {
                                auto col2 = cmap->get_color( p );

                                if ( red != col2.red ||
                                     green != col2.green ||
                                     blue != col2.blue ) {
                                    LOG_CRITICAL( "different calculated EHB color (%u,%u,%u) than cmap color (%u,%u,%u)",
                                                  red, green, blue,
                                                  col2.red, col2.green, col2.blue );
                                }
                            }
                        } else {
                            red = col.red;
                            green = col.green;
                            blue = col.blue;
                        }
                    } else {
                        auto col = cmap->get_color( p & 0xff );

                        if ( adjust_colors ) {
                            if ( ! ( bmhd->cmap_flags() & ChunkBMHD::CMAP_FLAG_8BIT ) ) {
                                col.red = ( col.red & 0xf0 ) | ( col.red >> 4 );
                                col.green = ( col.green & 0xf0 ) | ( col.green >> 4 );
                                col.blue = ( col.blue & 0xf0 ) | ( col.blue >> 4 );
                            }
                        }

                        red = col.red;
                        green = col.green;
                        blue = col.blue;
                    }
                }

                uint32_t rgba_value = p & 0xff000000;
                rgba_value |= ( blue & 0xff ) << 16;
                rgba_value |= ( green & 0xff ) << 8;
                rgba_value |= ( red & 0xff );
                rgba_pixels.push_back( rgba_value );
                pixels_written++;

                if ( ( pixels_written % pixels_per_line ) == 0 ) {
                    ham_hold_red = 0;
                    ham_hold_green = 0;
                    ham_hold_blue = 0;
                } else {
                    ham_hold_red = red;
                    ham_hold_green = green;
                    ham_hold_blue = blue;
                }
            } else {
                LOG_CRITICAL( "can't decode picture without color map" );
            }
        } else {
            int red = p & 0xff;
            int green = ( p >> 8 ) & 0xff;
            int blue = ( p >> 16 ) & 0xff;

            uint32_t rgba_value = p & 0xff000000;
            rgba_value |= ( blue & 0xff ) << 16;
            rgba_value |= ( green & 0xff ) << 8;
            rgba_value |= ( red & 0xff );
            rgba_pixels.push_back( rgba_value );
            pixels_written++;
        }
    }

    // output missing pixels as black pixels
    for ( size_t p = pixels_written; p < pixels_per_line * bmhd->height(); p++ ) {
        rgba_pixels.push_back( 0 );
    }

    return 0;
}

int ChunkFORM::convert_ilbm_to_rgba( const shared_properties_t &shared_props,
                                     std::vector< uint32_t > &rgba_pixels ) const
{
    const auto bmhd = m_ilbm_chunks.bmhd ? m_ilbm_chunks.bmhd : get_shared_chunk< ChunkBMHD >( "ILBM", "BMHD", shared_props );
    const auto cmap = m_ilbm_chunks.cmap ? m_ilbm_chunks.cmap : get_shared_chunk< ChunkCMAP >( "ILBM", "CMAP", shared_props );
    const auto camg = m_ilbm_chunks.camg ? m_ilbm_chunks.camg : get_shared_chunk< ChunkCAMG >( "ILBM", "CAMG", shared_props );

    std::vector< uint32_t > pixels;

    int res = convert_ilbm_to_pixels( shared_props,
                                      pixels );
    if ( res != 0 ) {
        return res;
    }

    res = convert_pixels_to_rgba( bmhd, cmap, camg, pixels, rgba_pixels );

    return res;
}

int ChunkFORM::dump_ilbm_as_binary( const shared_properties_t &shared_props,
                                    std::list< IFFBinaryResult > &output,
                                    bool include_raw ) const
{
    const auto bmhd = m_ilbm_chunks.bmhd ? m_ilbm_chunks.bmhd : get_shared_chunk< ChunkBMHD >( "ILBM", "BMHD", shared_props );

    if ( ! bmhd ) {
        LOG_CRITICAL( "no ILBM BMDH chunk" );
        return -EINVAL;
    }

    std::vector< uint32_t > rgba_pixels;
    int res = convert_ilbm_to_rgba( shared_props, rgba_pixels );

    if ( res != 0 ) {
        LOG_CRITICAL( "error decoding pixels: %d", res );
        return res;
    }

    LOG_DEBUG( "got %u pixel data", rgba_pixels.size() );

    return create_rgba_binary_result( rgba_pixels, bmhd->width(), bmhd->height(),
                                      bmhd->line_pitch(),
                                      output, "" );
}

int ChunkFORM::dump_deep_as_binary( const shared_properties_t &shared_props,
                                    std::list< IFFBinaryResult > &output,
                                    bool include_raw ) const
{
    const auto dgbl = m_deep_chunks.dgbl ? m_deep_chunks.dgbl : get_shared_chunk< ChunkDGBL >( "DEEP", "DGBL", shared_props );

    if ( ! dgbl ) {
        LOG_CRITICAL( "no DEEP DGBL chunk" );
        return -EINVAL;
    }

    const auto dbod = m_deep_chunks.dbod ? m_deep_chunks.dbod : get_shared_chunk< ChunkDBOD >( "DEEP", "DBOD", shared_props );

    if ( ! dbod ) {
        LOG_CRITICAL( "no DEEP DBOD chunk" );
        return -EINVAL;
    }

    const auto dloc = m_deep_chunks.dloc ? m_deep_chunks.dloc : get_shared_chunk< ChunkDLOC >( "DEEP", "DLOC", shared_props );

    if ( ! dloc ) {
        LOG_CRITICAL( "no DEEP DLOC chunk" );
        return -EINVAL;
    }

    const auto dpel = m_deep_chunks.dpel ? m_deep_chunks.dpel : get_shared_chunk< ChunkDPEL >( "DEEP", "DPEL", shared_props );

    if ( ! dpel ) {
        LOG_CRITICAL( "no DEEP DPEL chunk" );
        return -EINVAL;
    }

    std::vector< uint32_t > rgba_pixels;
    int res = dbod->decodeRGBA( dgbl, dpel, dloc, rgba_pixels );

    if ( res != 0 ) {
        LOG_CRITICAL( "error decoding pixels: %d", res );
        return res;
    }

    LOG_DEBUG( "got %u pixel data", rgba_pixels.size() );

    return create_rgba_binary_result( rgba_pixels, dloc->width(), dloc->height(),
                                      dloc->width(),
                                      output, "" );
}

int ChunkFORM::create_rgba_binary_result( const std::vector< uint32_t > &rgba_pixels,
                                          uint32_t width, uint32_t height, size_t line_pitch,
                                          std::list< IFFBinaryResult > &output,
                                          const std::string &name_info ) const
{
    IFFBinaryResult res1;
    res1.file_ending = "ppm";
    res1.file_name_info = "image-rgb";
    res1.main_file = true;

    if ( ! name_info.empty() ) {
        res1.file_name_info = name_info + "-" + res1.file_name_info;
    }

    res1.data.resize( 3 + 32 + 1 + 32 + 1 + 4 + 1 + width * height * 3 );
    size_t l = snprintf( &res1.data[0], res1.data.size(), "P6\n%u %u\n255\n", width, height );

    size_t pixels_written = 0;

    size_t pixels_per_line = line_pitch;
    size_t pixel_count = 0;
    
    for ( uint32_t p : rgba_pixels ) {
        if ( pixel_count == pixels_per_line ) {
            pixel_count = 0;
        }

        if ( pixel_count < width ) {
            int red = p & 0xff;
            int green = ( p >> 8 ) & 0xff;
            int blue = ( p >> 16 ) & 0xff;

            res1.data[l++] = (uint8_t)red;
            res1.data[l++] = (uint8_t)green;
            res1.data[l++] = (uint8_t)blue;
            pixels_written++;
        }

        pixel_count++;
    }

    // output missing pixels as black pixels
    for ( size_t p = pixels_written; p < width * height; p++ ) {
        res1.data[l++] = (uint8_t)0;
        res1.data[l++] = (uint8_t)0;
        res1.data[l++] = (uint8_t)0;
    }

    res1.data.resize( l );

    output.push_back( res1 );

    /* output the mask */
    IFFBinaryResult res2;
    res2.file_ending = "pgm";
    res2.file_name_info = "image-mask";

    if ( ! name_info.empty() ) {
        res2.file_name_info = name_info + "-" + res2.file_name_info;
    }

    res2.data.resize( 3 + 32 + 1 + 32 + 1 + 4 + 1 + width * height );
    l = snprintf( &res2.data[0], res2.data.size(), "P5\n%u %u\n255\n", width, height );

    pixels_written = 0;
    pixel_count = 0;

    for ( uint32_t p : rgba_pixels ) {
        if ( pixel_count == pixels_per_line ) {
            pixel_count = 0;
        }

        if ( pixel_count < width ) {
            uint8_t mask = p >> 24;

            res2.data[l++] = (uint8_t)mask;
            pixels_written++;
        }

        pixel_count++;
    }

    for ( size_t p = pixels_written; p < width * height; p++ ) {
        res2.data[l++] = (uint8_t)255;
    }

    res2.data.resize( l );

    output.push_back( res2 );

    return 0;
}

int ChunkFORM::dump_8svx_as_binary( const shared_properties_t &shared_props,
                                    std::list< IFFBinaryResult > &output,
                                    bool include_raw ) const
{
    const auto vhdr = m_8svx_chunks.vhdr ? m_8svx_chunks.vhdr : get_shared_chunk< ChunkVHDR >( "8SVX", "VHDR", shared_props );

    if ( ! vhdr ) {
        LOG_CRITICAL( "no 8SVX VHDR chunk" );
        return -EINVAL;
    }

    const auto body = m_8svx_chunks.body ? m_8svx_chunks.body : get_shared_chunk< Chunk8SVXBODY >( "8SVX", "BODY", shared_props );

    if ( ! body ) {
        LOG_CRITICAL( "no 8SVX BODY chunk" );
        return -EINVAL;
    }

    std::vector< std::vector< int8_t > > samples;

    int res = body->decode( vhdr, samples );
    if ( res != 0 ) {
        return res;
    }

    LOG_DEBUG( "got %u sample vectors", samples.size() );

    for ( size_t s = 0; s < samples.size(); s++ ) {
        std::vector< uint8_t > wav_data;

        int res = convert_8svx_samples_to_wav( vhdr, wav_data, samples[s] );
        if ( res != 0 ) {
            return res;
        }

        IFFBinaryResult res1;

        res1.file_ending = "wav";
        res1.main_file = false;

        if ( ( s % 2 ) == 0 ) {
            res1.file_name_info = "one_shot";
            if ( s == 0 ) {
                res1.main_file = true;
            }
        } else {
            res1.file_name_info = "repeat";
        }
        for ( auto ch : wav_data ) {
            res1.data.push_back( ch );
        }

        output.push_back( res1 );
    }

    return 0;
}

struct sf_user_data {
    std::vector< uint8_t > &wav_data;
    size_t pos = 0;
};

sf_count_t my_sf_get_filelen( void *user_data )
{
    struct sf_user_data *ud = (struct sf_user_data *)user_data;

    return ud->pos;
}

sf_count_t my_sf_seek( sf_count_t offset, int whence, void *user_data )
{
    struct sf_user_data *ud = (struct sf_user_data *)user_data;

    if ( whence == SEEK_CUR ) {
        ud->pos += offset;
    } else if ( whence == SEEK_SET ) {
        ud->pos = offset;
    } else if ( whence == SEEK_END ) {
        if ( offset > (sf_count_t)ud->wav_data.size() ) {
            ud->pos = 0;
        } else {
            ud->pos = ud->wav_data.size() - offset;
        }
    } else {
        LOG_CRITICAL( "invalid seek mode" );
        return -EINVAL;
    }

    return ud->pos;
}

sf_count_t my_sf_write( const void *ptr,
                        sf_count_t count,
                        void *user_data )
{
    struct sf_user_data *ud = (struct sf_user_data *)user_data;

    size_t l = ud->pos + count;
    if ( ud->wav_data.size() < l ) {
        ud->wav_data.resize( l );
    }
    std::copy_n( (const uint8_t*)ptr, count, &ud->wav_data[ ud->pos ]  );

    ud->pos += count;

    return count;
}

sf_count_t my_sf_tell( void *user_data )
{
    struct sf_user_data *ud = (struct sf_user_data *)user_data;

    return ud->pos;
}

int ChunkFORM::convert_8svx_samples_to_wav( std::shared_ptr< ChunkVHDR > vhdr,
                                            std::vector< uint8_t > &wav_data,
                                            const std::vector< int8_t > &samples ) const
{
    LOG_DEBUG( "called" );

    SF_VIRTUAL_IO sf_cb = {
                           .get_filelen = my_sf_get_filelen,
                           .seek = my_sf_seek,
                           .read = NULL,
                           .write = my_sf_write,
                           .tell = my_sf_tell
    };

    SF_INFO info = {
                    .frames  = (sf_count_t)samples.size(),
                    .samplerate = vhdr->samples_per_second(),
                    .channels = 1,
                    .format = SF_FORMAT_WAV | SF_FORMAT_PCM_16,
                    .sections = 0,
                    .seekable = 0
    };

    sf_user_data ud = { .wav_data = wav_data };

    SNDFILE *sfh = 	sf_open_virtual( &sf_cb, SFM_WRITE,
                                     &info, &ud);

    if ( ! sfh ) {
        LOG_CRITICAL( "could not create libsndfile handle: %s", sf_strerror( NULL ) );
        return -EFAULT;
    }

    for ( auto &s : samples ) {
        double v = s;

        if ( ! attribute( "8svx.ignore_volume", 0 ) ) {
            v *= vhdr->volume();
            v /= 0x10000;
        }
        v *= 256;
        short sv = (short)v;

        sf_count_t written = sf_write_short( sfh,
                                             &sv, 1 );
        if ( written != 1 ) {
            LOG_CRITICAL( "failed to write sample" );
            break;
        }
    }

    int res = sf_close( sfh );

    if ( res != 0 ) {
        LOG_CRITICAL( "failed to close libsndfile handle" );
        return -EFAULT;
    }

    return 0;
}

int ChunkFORM::dump_anim_as_binary( const shared_properties_t &shared_props,
                                    std::list< IFFBinaryResult > &output,
                                    bool include_raw ) const
{
    std::vector< std::vector< std::vector< uint8_t > > > planes_buffer;
    std::vector< uint8_t > mask;
    std::vector< uint32_t > pixels;
    std::vector< uint32_t > rgba_pixels;
    uint8_t output_buffer, delta_buffer;

    if ( m_anim_chunks.form_chunks.empty() ) {
        LOG_DEBUG( "no images within ANIM" );
        return 0;
    }

    auto chunk_it = m_anim_chunks.form_chunks.begin();
    const auto &form_chunk = *chunk_it;

    const auto bmhd = form_chunk->ilbm_bmhd();
    const auto body = form_chunk->ilbm_body();
    auto cmap = form_chunk->ilbm_cmap();
    const auto camg = form_chunk->ilbm_camg();
    const auto base_anhd = form_chunk->ilbm_camg();

    if ( ! bmhd ) {
        LOG_CRITICAL( "no BMHD chunk" );
        return -EINVAL;
    }
    if ( ! body ) {
        LOG_CRITICAL( "no BODY chunk" );
        return -EINVAL;
    }
    if ( ! cmap ) {
        LOG_CRITICAL( "no CMAP chunk" );
        return -EINVAL;
    }
    if ( ! camg ) {
        LOG_CRITICAL( "no CAMG chunk" );
        return -EINVAL;
    }

    planes_buffer.resize( 2 );

    int res = body->get_planes( bmhd,
                                planes_buffer[0],
                                mask );
    if ( res != 0 ) {
        return res;
    }

    res = body->planes_to_pixels( bmhd,
                                  planes_buffer[0],
                                  mask,
                                  pixels );
    if ( res != 0 ) {
        return res;
    }

    res = form_chunk->convert_pixels_to_rgba( bmhd, cmap, camg,
                                              pixels,
                                              rgba_pixels );
    if ( res != 0 ) {
        return res;
    }

    int frame_number = 0;
    std::string name_info;
    if ( base_anhd ) {
        // if the first ilbm has an anhd header, it's part of the animation
        name_info = formatStringToString( "frame-%05d", frame_number );
    } else {
        name_info = formatStringToString( "base_image" );
    }

    res = create_rgba_binary_result( rgba_pixels, bmhd->width(), bmhd->height(),
                                     bmhd->line_pitch(),
                                     output, name_info );
    if ( res != 0 ) {
        return res;
    }

    chunk_it++;

    output_buffer = 1;
    delta_buffer = 0;
    planes_buffer[1] = planes_buffer[0];

    uint32_t last_reltime = 0;

    for ( ; chunk_it != m_anim_chunks.form_chunks.end(); chunk_it++ ) {
        const auto anhd = (*chunk_it)->ilbm_anhd();
        const auto dlta = (*chunk_it)->ilbm_dlta();
        auto dlta_cmap = (*chunk_it)->ilbm_cmap();

        if ( ! anhd ) {
            continue;
        }

        if ( ! dlta ) {
            continue;
        }

        if ( dlta_cmap ) {
            cmap = dlta_cmap;
        }

        uint8_t target_buffer = anhd->interleave() == 1 ? output_buffer : delta_buffer;
        std::vector< std::vector< uint8_t > > new_planes;

        res = dlta->apply_delta( bmhd,
                                 anhd,
                                 planes_buffer[ target_buffer ],
                                 new_planes );
        if ( res != 0 ) {
            return res;
        }

        if ( anhd->reltime() > 1 ) {
            if ( attribute( "anim.duplicate_frames", 0 ) ) {
                for ( uint32_t repeat = anhd->reltime() - 1;
                      repeat > 0; repeat-- ) {
                    name_info = formatStringToString( "frame-%05d", frame_number++ );

                    res = create_rgba_binary_result( rgba_pixels, bmhd->width(), bmhd->height(),
                                                     bmhd->line_pitch(),
                                                     output, name_info );
                    if ( res != 0 ) {
                        return res;
                    }
                }
            }
        }

        planes_buffer[ target_buffer ] = new_planes;
        pixels.clear();
        rgba_pixels.clear();

        res = body->planes_to_pixels( bmhd,
                                      planes_buffer[ output_buffer ],
                                      mask,
                                      pixels );
        if ( res != 0 ) {
            return res;
        }

        res = form_chunk->convert_pixels_to_rgba( bmhd,
                                                  cmap,
                                                  camg,
                                                  pixels,
                                                  rgba_pixels );
        if ( res != 0 ) {
            return res;
        }

        if ( ! attribute( "anim.duplicate_frames", 0 ) ) {
            frame_number += anhd->reltime();
        }
        name_info = formatStringToString( "frame-%05d", frame_number );
        if ( attribute( "anim.duplicate_frames", 0 ) ) {
            frame_number++;
        }

        res = create_rgba_binary_result( rgba_pixels, bmhd->width(), bmhd->height(),
                                         bmhd->line_pitch(),
                                         output, name_info );
        if ( res != 0 ) {
            return res;
        }

        if ( anhd->interleave() == 1 ) {
        } else {
            delta_buffer = ( delta_buffer + 1 ) % 2;
            output_buffer = ( output_buffer + 1 ) % 2;
        }

        last_reltime = anhd->reltime();
    }

    // output last frame
    if ( output_buffer != delta_buffer ) {
        pixels.clear();
        rgba_pixels.clear();

        res = body->planes_to_pixels( bmhd,
                                      planes_buffer[ output_buffer ],
                                      mask,
                                      pixels );
        if ( res != 0 ) {
            return res;
        }

        res = form_chunk->convert_pixels_to_rgba( bmhd,
                                                  cmap,
                                                  camg,
                                                  pixels,
                                                  rgba_pixels );
        if ( res != 0 ) {
            return res;
        }

        if ( ! attribute( "anim.duplicate_frames", 0 ) ) {
            frame_number += last_reltime;
        }
        name_info = formatStringToString( "frame-%05d", frame_number++ );

        res = create_rgba_binary_result( rgba_pixels, bmhd->width(), bmhd->height(),
                                         bmhd->line_pitch(),
                                         output, name_info );
        if ( res != 0 ) {
            return res;
        }
    }

    return 0;
}

const std::shared_ptr< ChunkBMHD > &ChunkFORM::ilbm_bmhd() const
{
    return m_ilbm_chunks.bmhd;
}

const std::shared_ptr< ChunkBODY > &ChunkFORM::ilbm_body() const
{
    return m_ilbm_chunks.body;
}

const std::shared_ptr< ChunkCMAP > &ChunkFORM::ilbm_cmap() const
{
    return m_ilbm_chunks.cmap;
}

const std::shared_ptr< ChunkCAMG > &ChunkFORM::ilbm_camg() const
{
    return m_ilbm_chunks.camg;
}

const std::shared_ptr< ChunkANHD > &ChunkFORM::ilbm_anhd() const
{
    return m_ilbm_chunks.anhd;
}

const std::shared_ptr< ChunkDLTA > &ChunkFORM::ilbm_dlta() const
{
    return m_ilbm_chunks.dlta;
}

int ChunkFORM::parseKnownTypes( uint32_t form_type,
                                const char *buf,
                                size_t buflen )
{
    LOG_DEBUG( "called" );

    if ( buflen < 4 ) {
        return -EINVAL;
    }

    for ( const auto &elem : m_list_of_known_chunks ) {
        if ( CHUNKTYPE2NAME( form_type ) == elem.first ) {
            LOG_DEBUG( "found known type %s", elem.first.c_str() );

            buf += 4;
            buflen -= 4;

            while ( buflen >= 8 ) {
                //auto chunktype = betoh_u32( get_u32( buf, 0 ) );
                auto chunklen = betoh_u32( get_u32( buf, 4 ) );

                if ( chunklen + 8 > buflen ) {
                    LOG_DEBUG( "Invalid chunk length" );
                    return -EINVAL;
                }

                size_t chunk_parsed_len;

                auto chunk = std::make_shared< ChunkGeneric >( elem.second );
                if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                    LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                    if ( fail_on_unsupported() ) {
                        return -EINVAL;
                    }
                }

                m_generic_chunks.push_back( chunk );

                if ( chunklen & 1 ) {
                    buf += 8 + chunklen + 1;
                    buflen -= 8 + chunklen + 1;
                } else {
                    buf += 8 + chunklen;
                    buflen -= 8 + chunklen;
                }
            }

            return 0;
        }
    }

    return -ENOTSUP;
}

bool ChunkFORM::isKnownType( uint32_t form_type ) const
{
    for ( const auto &elem : m_list_of_known_chunks ) {
        if ( CHUNKTYPE2NAME( form_type ) == elem.first ) {
            return true;
        }
    }

    return false;
}

int ChunkFORM::fill_info( IFFInfo &info ) const
{
    info = m_info;
    info.set_file_type( "FORM " + CHUNKTYPE2NAME( m_form_type ) );

    if ( m_form_type == CHUNKNAME( "ILBM" ) ) {
        const auto camg = m_ilbm_chunks.camg;
        const auto bmhd = m_ilbm_chunks.bmhd;

        if ( bmhd ) {
            bool ham = false;
            bool ehb = false;
            if ( camg ) {
                ham = camg->display_mode() & ChunkCAMG::HAM;
                ehb = camg->display_mode() & ChunkCAMG::EXTRA_HALFBRITE;
            }
            bool ham8 = ham && bmhd->nplanes() == 8;

            std::stringstream ss1;

            ss1 << bmhd->width() << "x" << bmhd->height();

            if ( bmhd->nplanes() == 6 && ham ) {
                ss1 << " (" << ( 1 << 12 ) << " colors (HAM))";
            } else if ( bmhd->nplanes() == 8 && ham8 ) {
                ss1 << " (" << ( 1 << 24 ) << " colors (HAM8))";
            } else if ( bmhd->nplanes() == 6 && ehb ) {
                ss1 << " (" << ( 1 << 6 ) << " colors (EHB))";
            } else {
                ss1 << " (" << ( 1 << bmhd->nplanes() ) << " colors)";
            }

            info.set_additional_info( ss1.str() );
        }
    }

    return 0;
}
