/* chunk_form.hh
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CHUNK_FORM_HH
#define CHUNK_FORM_HH

#include "chunk_generic.hh"

class ChunkBMHD;
class ChunkCMAP;
class ChunkGRAB;
class ChunkDEST;
class ChunkSPRT;
class ChunkCAMG;
class ChunkBODY;
class ChunkANNO;
class ChunkDPI;
class ChunkCRNG;

class ChunkDGBL;
class ChunkDPEL;
class ChunkDLOC;
class ChunkDBOD;
class ChunkDCHG;

class ChunkVHDR;
class ChunkNAME;
class Chunk8SVXBODY;

class ChunkANHD;
class ChunkDLTA;

class ChunkFHDR;

class ChunkFORM : public ChunkGeneric {
public:
    std::string dump_as_readable_text( const shared_properties_t &shared_props ) override;
    json_t *dump_as_json( const shared_properties_t &shared_props,
                          bool include_raw ) const override;
    int dump_as_binary( const shared_properties_t &shared_props,
                        std::list< IFFBinaryResult > &output,
                        bool include_raw ) const override;
    int fill_info( IFFInfo &info ) const override;

    const std::shared_ptr< ChunkBMHD > &ilbm_bmhd() const;
    const std::shared_ptr< ChunkBODY > &ilbm_body() const;
    const std::shared_ptr< ChunkCMAP > &ilbm_cmap() const;
    const std::shared_ptr< ChunkCAMG > &ilbm_camg() const;
    const std::shared_ptr< ChunkANHD > &ilbm_anhd() const;
    const std::shared_ptr< ChunkDLTA > &ilbm_dlta() const;
protected:
    int parse_chunk( const std::vector< char > &data ) override;
private:
    int parseGeneric( const char *buf,
                      size_t buflen );
    int parseFTXT( const char *buf,
                   size_t buflen );
    int parseWOWO( const char *buf,
                   size_t buflen );
    int parseILBM( const char *buf,
                   size_t buflen );
    int parseDEEP( const char *buf,
                   size_t buflen );
    int parseSWRT( const char *buf,
                   size_t buflen );
    int parseFANT( const char *buf,
                   size_t buflen );
    int parse8SVX( const char *buf,
                   size_t buflen );
    int parseANIM( const char *buf,
                   size_t buflen );
    int parsePREF( const char *buf,
                   size_t buflen );
    int parseCTLG( const char *buf,
                   size_t buflen );
    int parseCDVR( const char *buf,
                   size_t buflen );
    int parseREFL( const char *buf,
                   size_t buflen );
    int parseGXUI( const char *buf,
                   size_t buflen );
    int parseSMUS( const char *buf,
                   size_t buflen );
    int parseGXWD( const char *buf,
                   size_t buflen );
    int parseKICK( const char *buf,
                   size_t buflen );
    int parseKnownTypes( uint32_t form_type,
                         const char *buf,
                         size_t buflen );

    bool isKnownType( uint32_t form_type ) const;

    int dump_ilbm_as_binary( const shared_properties_t &shared_props,
                             std::list< IFFBinaryResult > &output,
                             bool include_raw ) const;
    int convert_ilbm_to_pixels( const shared_properties_t &shared_props,
                                std::vector< uint32_t > &pixels ) const;
    static int convert_pixels_to_rgba( const std::shared_ptr< ChunkBMHD > &bmhd,
                                       const std::shared_ptr< ChunkCMAP > &cmap,
                                       const std::shared_ptr< ChunkCAMG > &camg,
                                       const std::vector< uint32_t > &pixels,
                                       std::vector< uint32_t > &rgba_pixels );
    int convert_ilbm_to_rgba( const shared_properties_t &shared_props,
                              std::vector< uint32_t > &rgba_pixels ) const;

    int dump_deep_as_binary( const shared_properties_t &shared_props,
                             std::list< IFFBinaryResult > &output,
                             bool include_raw ) const;

    int create_rgba_binary_result( const std::vector< uint32_t > &rgba_pixels,
                                   uint32_t width, uint32_t height, size_t line_pitch,
                                   std::list< IFFBinaryResult > &output,
                                   const std::string &name_info ) const;

    int dump_8svx_as_binary( const shared_properties_t &shared_props,
                             std::list< IFFBinaryResult > &output,
                             bool include_raw ) const;
    int convert_8svx_samples_to_wav( std::shared_ptr< ChunkVHDR > vhdr,
                                     std::vector< uint8_t > &wav_data,
                                     const std::vector< int8_t > &samples ) const;

    int dump_anim_as_binary( const shared_properties_t &shared_props,
                             std::list< IFFBinaryResult > &output,
                             bool include_raw ) const;

    uint32_t m_form_type = 0;

    const static std::list< std::pair< std::string, std::list< std::string > > > m_list_of_known_chunks;

    IFFInfo m_info;

    std::list< std::shared_ptr< ChunkGeneric > > m_generic_chunks;

    std::list< std::shared_ptr< ChunkGeneric > > m_ftxt_chunks;
    std::list< std::shared_ptr< ChunkGeneric > > m_wowo_chunks;

    struct {
        std::shared_ptr< ChunkBMHD > bmhd;
        std::shared_ptr< ChunkCMAP > cmap;
        std::shared_ptr< ChunkGRAB > grab;
        std::shared_ptr< ChunkDEST > dest;
        std::shared_ptr< ChunkSPRT > sprt;
        std::shared_ptr< ChunkCAMG > camg;
        std::shared_ptr< ChunkBODY > body;
        std::shared_ptr< ChunkANNO > anno;
        std::shared_ptr< ChunkDPI > dpi;
        std::shared_ptr< ChunkCRNG > crng;
        std::shared_ptr< ChunkANHD > anhd;
        std::shared_ptr< ChunkDLTA > dlta;
        std::list< std::shared_ptr< ChunkGeneric > > known_chunks;
        std::list< std::shared_ptr< ChunkGeneric > > other_chunks;
    } m_ilbm_chunks;

    struct {
        std::shared_ptr< ChunkDGBL > dgbl;
        std::shared_ptr< ChunkDPEL > dpel;
        std::shared_ptr< ChunkDLOC > dloc;
        std::shared_ptr< ChunkDBOD > dbod;
        std::shared_ptr< ChunkDCHG > dchg;
        std::list< std::shared_ptr< ChunkGeneric > > known_chunks;
        std::list< std::shared_ptr< ChunkGeneric > > other_chunks;
    } m_deep_chunks;

    std::list< std::shared_ptr< ChunkGeneric > > m_swrt_chunks;

    struct {
        std::shared_ptr< ChunkFHDR > fhdr;
        std::list< std::shared_ptr< ChunkGeneric > > chunks;
    } m_fant_chunks;

    struct {
        std::shared_ptr< ChunkVHDR > vhdr;
        std::shared_ptr< ChunkANNO > anno;
        std::shared_ptr< ChunkNAME > name;
        std::shared_ptr< Chunk8SVXBODY > body;
        std::list< std::shared_ptr< ChunkGeneric > > other_chunks;
    } m_8svx_chunks;

    struct {
        std::list< std::shared_ptr< ChunkFORM > > form_chunks;
        std::list< std::shared_ptr< ChunkGeneric > > other_chunks;
    } m_anim_chunks;

    std::list< std::shared_ptr< ChunkGeneric > > m_pref_chunks;

    std::list< std::shared_ptr< ChunkGeneric > > m_ctlg_chunks;

    std::list< std::shared_ptr< ChunkGeneric > > m_cdvr_chunks;

    std::list< std::shared_ptr< ChunkGeneric > > m_refl_chunks;

    std::list< std::shared_ptr< ChunkGeneric > > m_gxui_chunks;

    std::list< std::shared_ptr< ChunkGeneric > > m_smus_chunks;

    std::list< std::shared_ptr< ChunkGeneric > > m_opus_chunks;
};

#endif /* CHUNK_FORM_HH */
