/* chunk_generic.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_generic.hh"
#include <algorithm>
#include <ctype.h>

bool ChunkGeneric::s_fail_on_unsupported = false;
bool ChunkGeneric::s_try_known_chunks = false;
std::map< std::string, int > ChunkGeneric::s_attributes;

ChunkGeneric::~ChunkGeneric()
{}

int ChunkGeneric::parse( const char *buf,
                         size_t buflen, size_t &parsed_len )
{
    int res = read_into_raw( buf, buflen, parsed_len );

    if ( res != 0 ) {
        return res;
    }

    return parse_chunk( m_raw_data );
}

int ChunkGeneric::parse_chunk( const std::vector< char > &data )
{
    if ( ! m_accepted_types.empty() ) {
        bool found = false;
        for ( auto &type : m_accepted_types ) {
            if ( type.size() == 4 ) {
                LOG_DEBUG( "test type %s", type.c_str() );
                if ( m_chunk_type == CHUNKNAME( type ) ) {
                    found = true;
                    break;
                }
            }
        }

        if ( ! found ) {
            LOG_DEBUG( "not in accepted types" );
            return -EINVAL;
        }
    }

    LOG_DEBUG( "accept type" );

    return 0;
}

int ChunkGeneric::read_into_raw( const char *buf,
                                 size_t buflen,
                                 size_t &raw_chunklen )
{
    if ( buflen < 8 ) {
        LOG_DEBUG( "chunk too small" );
        return -EINVAL;
    }

    m_chunk_type = betoh_u32( get_u32( buf, 0 ) );

    auto chunklen = betoh_u32( get_u32( buf, 4 ) );

    if ( chunklen + 8 > buflen ||
         chunklen < 0 ) {
        LOG_DEBUG( "chunk smaller than announced" );
        return -EINVAL;
    }

    m_raw_data.resize( chunklen );
    std::copy_n( buf + 8, chunklen,
                 m_raw_data.begin() );

    if ( chunklen & 1 ) {
        raw_chunklen = 8 + chunklen + 1;
    } else {
        raw_chunklen = 8 + chunklen;
    }

    LOG_DEBUG( "Read chunk '%.*s' of %u bytes into raw_data for parsing", 4, buf, chunklen );

    return 0;
}

json_t *ChunkGeneric::dump_as_json( const shared_properties_t &shared_props, bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( CHUNKTYPE2NAME( m_chunk_type ).c_str() ) );

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

int ChunkGeneric::dump_raw_to_json( json_t *elem ) const
{
    auto jarray = json_array();
    std::string raw_ascii;
    for ( auto &ch : m_raw_data ) {
        json_array_append_new( jarray, json_integer( (uint8_t)ch ) );
        if ( isprint( ch ) ) {
            raw_ascii += ch;
        } else {
            raw_ascii += ' ';
        }
    }
    json_object_set_new( elem, "rawData", jarray );
    json_object_set_new( elem, "rawAscii", json_string( raw_ascii.c_str() ) );

    return 0;
}

void ChunkGeneric::set_fail_on_unsupported( bool v )
{
    s_fail_on_unsupported = v;
}

bool ChunkGeneric::fail_on_unsupported()
{
    return s_fail_on_unsupported;
}

void ChunkGeneric::set_try_known_chunks( bool v )
{
    s_try_known_chunks = v;
}

bool ChunkGeneric::try_known_chunks()
{
    return s_try_known_chunks;
}

std::string ChunkGeneric::dump_as_readable_text( const shared_properties_t &shared_props )
{
    return "";
}

std::string ChunkGeneric::getChunkName() const
{
    return CHUNKTYPE2NAME( m_chunk_type );
}

int ChunkGeneric::dump_as_binary( const shared_properties_t &shared_props,
                                  std::list< IFFBinaryResult > &output,
                                  bool include_raw ) const
{
    LOG_DEBUG( "called" );

    if ( include_raw ) {
        IFFBinaryResult res;

        res.file_ending = "raw";

        int raw_res = dump_as_raw( shared_props, res.data ); 
        if ( raw_res != 0 ) {
            return raw_res;
        }

        output.push_back( res );
    }

    return 0;
}

int ChunkGeneric::dump_as_raw( const shared_properties_t &shared_props,
                               std::vector< char > &data ) const
{
    data = m_raw_data;

    return 0;
}

std::pair< int, bool > ChunkGeneric::attribute( const std::string &key )
{
    auto it = s_attributes.find( key );
    if ( it == s_attributes.end() ) {
        return std::make_pair( 0, false );
    }

    return std::make_pair( it->second, true );
}

int ChunkGeneric::attribute( const std::string &key, int default_value )
{
    auto it = s_attributes.find( key );
    if ( it == s_attributes.end() ) {
        return default_value;
    }

    return it->second;
}

void ChunkGeneric::set_attribute( const std::string &key, int value )
{
    s_attributes[ key ] = value;
}

std::map< std::string, int > ChunkGeneric::get_attributes()
{
    return s_attributes;
}

int ChunkGeneric::fill_info( IFFInfo &info ) const
{
    return -ENOTSUP;
}

std::string IFFInfo::file_type() const
{
    return m_file_type;
}

void IFFInfo::set_file_type( const std::string &ft )
{
    m_file_type = ft;
}

std::string IFFInfo::description() const
{
    return m_description;
}

void IFFInfo::set_description( const std::string &descr )
{
    m_description = descr;
}

std::string IFFInfo::additional_info() const
{
    return m_additional_info;
}

void IFFInfo::set_additional_info( const std::string &info )
{
    m_additional_info = info;
}

std::string IFFInfo::get_info_as_text() const
{
    std::string res;

    res += "FILETYPE: ";
    res += file_type();
    res += "\n";
    res += "DESCRIPTION: ";
    res += description();
    res += "\n";
    res += "ADDITIONALINFO: ";
    res += additional_info();

    return res;
}

json_t *IFFInfo::get_info_as_json() const
{
    auto jinfo = json_object();
    json_object_set_new( jinfo, "filetype", json_string( file_type().c_str() ) );
    json_object_set_new( jinfo, "description", json_string( description().c_str() ) );
    json_object_set_new( jinfo, "additionalInfo", json_string( additional_info().c_str() ) );
    return jinfo;
}
