/* chunk_generic.hh
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CHUNK_GENERIC_H
#define CHUNK_GENERIC_H

#include "iff_convert.hh"

#include "lowlevel_types.hh"
#include <memory>
#include "ab_logging.hh"
#include <cstring>
#include <map>
#include <jansson.h>
#include <list>
#include <vector>

#define CHUNKNAME(name) ( (uint32_t)((name)[0]) << 24 | (uint32_t)((name)[1]) << 16 | (uint32_t)((name)[2]) << 8 | (uint32_t)((name)[3]) )
#define CHUNKTYPE2NAME(type) ( std::string() + (char)( (type) >> 24 ) + (char)( ( (type) >> 16 ) & 0xff ) + (char)( ( (type) >> 8 ) & 0xff ) + (char)( (type) &0xff ) )

class IFFInfo {
public:
    std::string file_type() const;
    void set_file_type( const std::string &ft );

    std::string description() const;
    void set_description( const std::string &descr );

    std::string additional_info() const;
    void set_additional_info( const std::string &info );

    std::string get_info_as_text() const;
    json_t *get_info_as_json() const;
private:
    std::string m_file_type;
    std::string m_description;
    std::string m_additional_info;
};

class ChunkGeneric {
public:
    ChunkGeneric()
    {}

    ChunkGeneric( const std::list< std::string > &accepted_types ) :
        m_accepted_types( accepted_types )
    {}

    virtual ~ChunkGeneric();

    virtual int parse( const char *buf,
                       size_t buflen,
                       size_t &parsed_len );

    virtual std::string getChunkName() const;

    class IFFBinaryResult {
    public:
        std::string file_name_info;
        std::string file_ending;
        std::vector< char > data;
        std::string temp_file;
        bool main_file = false;
    };

    typedef std::map< std::pair< std::string, std::string >, std::shared_ptr< ChunkGeneric > > shared_properties_t;

    virtual std::string dump_as_readable_text( const shared_properties_t &shared_props );
    virtual json_t *dump_as_json( const shared_properties_t &shared_props, bool include_raw ) const;
    virtual int dump_as_raw( const shared_properties_t &shared_props, std::vector< char > &data ) const;
    virtual int dump_as_binary( const shared_properties_t &shared_props,
                                std::list< IFFBinaryResult > &output,
                                bool include_raw ) const;
    virtual int dump_raw_to_json( json_t *elem ) const;

    virtual int fill_info( IFFInfo &info ) const;

    template< class CHUNKTYPE > std::shared_ptr< CHUNKTYPE > reparse()
    {
        std::shared_ptr< CHUNKTYPE > res = std::make_shared< CHUNKTYPE >();

        LOG_DEBUG( "reparse" );

        char buf[8 + m_raw_data.size()];
        uint32_t *ibuf = (uint32_t*)buf;
        ibuf[0] = htobe_u32( CHUNKNAME( getChunkName() ) );
        ibuf[1] = htobe_u32( m_raw_data.size() );
        memcpy( &buf[8], &m_raw_data[0], m_raw_data.size() );

        LOG_DEBUG( "Create new chunk '%.*s' for parsing", 4, buf );

        size_t parsed_len;
        int parse_res = res->parse( buf, sizeof( buf ), parsed_len );

        LOG_DEBUG( "parse result: %d", parse_res );

        if ( parse_res == 0 ) {
            return res;
        } else {
            return nullptr;
        }
    }

    template< class CHUNKTYPE > std::shared_ptr< CHUNKTYPE > get_shared_chunk( const std::string &form_type,
                                                                               const std::string &chunk_id,
                                                                               const shared_properties_t &shared_props ) const
    {
        auto prop_key = std::make_pair( form_type, chunk_id );
        auto it = shared_props.find( prop_key );
        if ( it != shared_props.end() ) {
            return it->second->reparse< CHUNKTYPE >();
        }

        return nullptr;
    }

    static void set_fail_on_unsupported( bool v );
    static bool fail_on_unsupported();

    static void set_try_known_chunks( bool v );
    static bool try_known_chunks();

    static std::pair< int, bool > attribute( const std::string &key );
    static int attribute( const std::string &key, int default_value );
    static void set_attribute( const std::string &key, int value );
    static std::map< std::string, int > get_attributes();
protected:
    virtual int read_into_raw( const char *buf,
                               size_t buflen,
                               size_t &raw_chunklen );
    virtual int parse_chunk( const std::vector< char > &data );

    uint32_t m_chunk_type = 0;
    std::vector< char > m_raw_data;
private:
    static bool s_fail_on_unsupported;
    static bool s_try_known_chunks;
    static std::map< std::string, int > s_attributes;

    std::list< std::string > m_accepted_types;
};

#endif /* CHUNK_GENERIC_H */
