/* chunk_list.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_list.hh"
#include "chunk_form.hh"
#include "chunk_cat.hh"
#include "chunk_prop.hh"
#include "find_chunk.hh"
#include <sstream>

int ChunkLIST::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();
    const char *buf = &data[0];

    if ( chunklen < 12 ) {
        return -EINVAL;
    }

    if ( m_chunk_type != CHUNKNAME( "LIST" ) ) {
        return -EINVAL;
    }

    LOG_DEBUG( "Found LIST chunk" );
    LOG_DEBUG( "parse %u bytes...", chunklen );

    m_list_type = betoh_u32( get_u32( buf, 0 ) );

    buf += 4;
    chunklen -= 4;

    while ( chunklen >= 8 ) {
        size_t inner_parsed_len = 0;

        auto elem = check_and_parse( buf, chunklen, inner_parsed_len );

        if ( ! elem ) {
            LOG_DEBUG( "Failed to parse LIST entry" );
            return -EINVAL;
        }

        auto prop_chunk = std::dynamic_pointer_cast< ChunkPROP >( elem );
        auto form_chunk = std::dynamic_pointer_cast< ChunkFORM >( elem );
        auto list_chunk = std::dynamic_pointer_cast< ChunkLIST >( elem );
        auto cat_chunk = std::dynamic_pointer_cast< ChunkCAT >( elem );
        if ( prop_chunk ) {
            if ( ! m_other_chunks.empty() ) {
                LOG_CRITICAL( "found PROP chunk after some other CHUNKS, not valid" );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }
            m_prop_chunks.push_back( prop_chunk );
        } else if ( form_chunk ||
                    list_chunk ||
                    cat_chunk ) {
            m_other_chunks.push_back( elem );
        } else {
            LOG_CRITICAL( "invalid chunk fond in LIST chunk" );
            if ( fail_on_unsupported() ) {
                return -EINVAL;
            }
            m_other_chunks.push_back( elem );
        }

        buf += inner_parsed_len;
        chunklen -= inner_parsed_len;
    }

    return 0;
}

std::string ChunkLIST::dump_as_readable_text( const shared_properties_t &shared_props )
{
    std::string res;

    auto new_props = add_prop_chunks( shared_props );

    for ( auto &chunk : m_other_chunks ) {
        res += chunk->dump_as_readable_text( new_props );
    }

    return res;
}

json_t *ChunkLIST::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG("called");

    auto new_props = add_prop_chunks( shared_props );

    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "LIST" ) );

    json_object_set_new( jroot, "subtype", json_string( CHUNKTYPE2NAME( m_list_type ).c_str() ) );

    auto jarray = json_array();

    for ( auto &chunk : m_other_chunks ) {
        auto jentry = chunk->dump_as_json( new_props, include_raw );

        json_array_append_new( jarray, jentry );
    }

    json_object_set_new( jroot, "content", jarray );

    return jroot;
}

int ChunkLIST::dump_as_binary( const shared_properties_t &shared_props,
                               std::list< IFFBinaryResult > &output,
                               bool include_raw ) const
{
    auto new_props = add_prop_chunks( shared_props );

    for ( auto &chunk : m_other_chunks ) {
        if ( chunk->dump_as_binary( new_props,
                                    output,
                                    include_raw ) != 0 ) {
            LOG_DEBUG( "failed to dump binary data" );
        }
    }

    return 0;
}

ChunkGeneric::shared_properties_t ChunkLIST::add_prop_chunks( const shared_properties_t &shared_props ) const
{
    shared_properties_t res = shared_props;

    std::string form_type = CHUNKTYPE2NAME( m_list_type );

    for ( const auto &e : m_prop_chunks ) {
        e->extend_shared_props( res, form_type );
    }

    return res;
}

int ChunkLIST::fill_info( IFFInfo &info ) const
{
    std::stringstream ss1;

    info.set_file_type( "LIST " + CHUNKTYPE2NAME( m_list_type ) );
    std::string descr = "List of ";
    std::string addinfo = "";

    int res = -ENOTSUP;

    for ( auto &chunk : m_other_chunks ) {
        IFFInfo sinfo;
        if ( chunk->fill_info( sinfo ) == 0 ) {
            if ( ! sinfo.description().empty() ) {
                descr += sinfo.description() + ";";
            }
            if ( ! sinfo.additional_info().empty() ) {
                addinfo += sinfo.additional_info() + ";";
            }
            res = 0;
        }
    }

    info.set_description( descr );
    info.set_additional_info( addinfo );

    return res;
}
