/* chunk_ssa_form.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_ssa_form.hh"
#include "chunk_form.hh"
#include <algorithm>

int ChunkSSAFORM::read_into_raw( const char *buf,
                                 size_t buflen,
                                 size_t &raw_chunklen )
{
    if ( buflen < 8 ) {
        return -EINVAL;
    }

    m_chunk_type = betoh_u32( get_u32( buf, 0 ) );

    auto chunklen = betoh_u32( get_u32( buf, 4 ) );

    m_ssa_variant = ChunkGeneric::attribute( "ssa.variant", 0 );

    // there seems to be a bug in the form handling so the chunklen
    // does not include the form type.

    if ( m_ssa_variant == 2 ) {
        chunklen += 4;
    } else if ( m_ssa_variant == 1 ) {
        chunklen += 4 + 8;
    } else {
        LOG_VERBOSE( "unknown variant of SSA file, trying to continue" );
    }

    if ( chunklen + 8 > buflen ||
         chunklen < 0 ) {
        LOG_DEBUG( "invalid chunklen %u/%u", chunklen, buflen );
        return -EINVAL;
    }

    m_raw_data.resize( chunklen );
    std::copy_n( buf + 8, chunklen,
                 m_raw_data.begin() );

    if ( chunklen & 1 ) {
        raw_chunklen = 8 + chunklen + 1;
    } else {
        raw_chunklen = 8 + chunklen;
    }

    LOG_DEBUG( "Read chunk '%.*s' of %u bytes into raw_data for parsing", 4, buf, chunklen );

    return 0;
}

int ChunkSSAFORM::parse_chunk( const std::vector< char > &data )
{
    auto buflen = data.size();
    const char *buf = &data[0];

    if ( buflen < 4 ) {
        LOG_DEBUG( "too short for SSA FORM" );
        return -EINVAL;
    }

    if ( m_chunk_type != CHUNKNAME( "FORM" ) ) {
        LOG_DEBUG( "not a SSA FORM chunk" );
        return -EINVAL;
    }

    LOG_DEBUG( "Found SSA FORM chunk" );
    LOG_DEBUG( "parse %u bytes...", buflen );

    m_ssaform_type = betoh_u32( get_u32( buf, 0 ) );
    auto ssaform_string = std::string( buf, 4 );

    LOG_DEBUG( "Found SSA FORM type %s", ssaform_string.c_str() );
    
    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        uint8_t extra_bytes = 0;
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        LOG_DEBUG( "Next chunk %.*s", 4, buf );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length: %u/%u", chunklen, buflen );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "FORM" ) ) {
            LOG_DEBUG( "Got FORM chunk of %u bytes", chunklen );
            if ( m_ssa_variant == 2 ) {
                // in variant 2, all FORM chunks have a invalid
                // chunklen of 4 less than the actual size
                auto form = std::make_shared< ChunkSSAFORM >();
                if ( form->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                    return -EFAULT;
                }
                m_chunks.push_back( form );

                extra_bytes = 4;
            } else {
                // in variant 1, only the outer FORM chunk has an
                // invalid lenght, all internal seem to be correct
                auto form = std::make_shared< ChunkFORM >();
                if ( form->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                    return -EFAULT;
                }
                m_chunks.push_back( form );
            }
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{} );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            LOG_DEBUG( "got generic chunk %.*s of %u bytes", 4, buf, chunk_parsed_len );

            m_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
        LOG_DEBUG( "skip %u extra bytes", extra_bytes );
        if ( extra_bytes > buflen ) {
            LOG_CRITICAL( "no %u extra bytes available", extra_bytes );
            break;
        }

        buf += extra_bytes;
        buflen -= extra_bytes;
    }

    LOG_DEBUG( "Done parsing SSA FORM chunk" );

    return 0;
}

std::string ChunkSSAFORM::dump_as_readable_text( const shared_properties_t &shared_props )
{
    std::string res;

    for ( auto &chunk : m_chunks ) {
        res += chunk->dump_as_readable_text( shared_props );
    }

    return res;
}

json_t *ChunkSSAFORM::dump_as_json( const shared_properties_t &shared_props,
                                    bool include_raw ) const
{
    LOG_DEBUG("called");

    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "FORM" ) );

    json_object_set_new( jroot, "subtype", json_string( CHUNKTYPE2NAME( m_ssaform_type ).c_str() ) );

    auto jarray = json_array();

    for ( auto &chunk : m_chunks ) {
        auto jentry = chunk->dump_as_json( shared_props, include_raw );

        json_array_append_new( jarray, jentry );
    }

    json_object_set_new( jroot, "content", jarray );

    return jroot;
}

int ChunkSSAFORM::dump_as_binary( const shared_properties_t &shared_props,
                               std::list< IFFBinaryResult > &output,
                               bool include_raw ) const
{
    std::string parsed_text;
    std::vector< char > raw_data;
    for ( auto &chunk : m_chunks ) {
        parsed_text += chunk->dump_as_readable_text( shared_props );

        if ( include_raw ) {
            std::vector< char > tdata;
            if ( chunk->dump_as_raw( shared_props, tdata ) == 0 ) {
                for ( auto ch : tdata ) {
                    raw_data.push_back( ch );
                }
            } else {
                LOG_CRITICAL( "failed to dump raw data" );
            }
        }
    }

    IFFBinaryResult res1;
    IFFBinaryResult res2;

    res1.file_ending = "txt";
    for ( auto ch : parsed_text ) {
        res1.data.push_back( ch );
    }

    output.push_back( res1 );

    if ( include_raw ) {
        res2.file_ending = "raw";
        res2.data = raw_data;

        output.push_back( res2 );
    }

    return 0;
}
