/* chunk_vhdr.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_vhdr.hh"
#include <sstream>

int ChunkVHDR::parse_chunk( const std::vector< char > &data )
{
    auto chunklen = data.size();

    if ( m_chunk_type != CHUNKNAME( "VHDR" ) ) {
        return -EINVAL;
    }

    const char *buf = &data[0];

    if ( chunklen != 20 ) {
        LOG_DEBUG( "invalid VHDR chunk size: %u", chunklen );
        return -EINVAL;
    }

    m_one_shot_samples = betoh_u32( get_u32( buf, 0 ) );
    m_repeat_samples = betoh_u32( get_u32( buf, 4 ) );
    m_samples_per_cycle = betoh_u32( get_u32( buf, 8 ) );
    m_samples_per_second = betoh_u16( get_u16( buf, 12 ) );
    m_no_of_octaves = get_u8( buf, 14 );
    m_compression = get_u8( buf, 15 );
    m_volume = (int32_t)betoh_u32( get_u32( buf, 16 ) );

    return 0;
}

json_t *ChunkVHDR::dump_as_json( const shared_properties_t &shared_props,
                                 bool include_raw ) const
{
    LOG_DEBUG( "called" );
    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "VHDR" ) );
    json_object_set_new( jroot, "oneShotSamples", json_integer( m_one_shot_samples ) );
    json_object_set_new( jroot, "repeatSamples", json_integer( m_repeat_samples ) );
    json_object_set_new( jroot, "samplesPerCycle", json_integer( m_samples_per_cycle ) );
    json_object_set_new( jroot, "samplesPerSecond", json_integer( m_samples_per_second ) );
    json_object_set_new( jroot, "noOfOctaves", json_integer( m_no_of_octaves ) );
    json_object_set_new( jroot, "volume", json_integer( m_volume ) );

    json_object_set_new( jroot, "compressionValue", json_integer( m_compression ) );
    switch ( m_compression ) {
        case COMPRESSION_NONE:
            json_object_set_new( jroot, "compression", json_string( "none" ) );
            break;
        case COMPRESSION_FDE:
            json_object_set_new( jroot, "compression", json_string( "fibonacciDeltaEncoding" ) );
            break;
        default:
            json_object_set_new( jroot, "compression", json_string( "invalid" ) );
            break;
    }

    if ( include_raw ) {
        dump_raw_to_json( jroot );
    }

    return jroot;
}

uint32_t ChunkVHDR::one_shot_samples() const
{
    return m_one_shot_samples;
}

uint32_t ChunkVHDR::repeat_samples() const
{
    return m_repeat_samples;
}

uint8_t ChunkVHDR::no_of_octaves() const
{
    return m_no_of_octaves;
}

uint8_t ChunkVHDR::compression() const
{
    return m_compression;
}

uint16_t ChunkVHDR::samples_per_second() const
{
    return m_samples_per_second;
}

int32_t ChunkVHDR::volume() const
{
    return m_volume;
}

int ChunkVHDR::fill_info( IFFInfo &info ) const
{
    std::stringstream ss1;

    ss1 << m_one_shot_samples + m_repeat_samples << " samples";

    info.set_additional_info( ss1.str() );

    return 0;
}
