/* chunk_wowo_form.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "chunk_wowo_form.hh"
#include "chunk_wtxt.hh"
#include "chunk_wpar.hh"

int ChunkWOWOFORM::parse_chunk( const std::vector< char > &data )
{
    auto buflen = data.size();
    const char *buf = &data[0];

    if ( buflen < 4 ) {
        LOG_DEBUG( "too short for WOWOFORM" );
        return -EINVAL;
    }

    if ( m_chunk_type != CHUNKNAME( "FORM" ) ) {
        LOG_DEBUG( "not a WOWOFORM chunk" );
        return -EINVAL;
    }

    LOG_DEBUG( "Found WOWO FORM chunk" );
    LOG_DEBUG( "parse %u bytes...", buflen );

    m_wowoform_type = betoh_u32( get_u32( buf, 0 ) );
    auto wowoform_string = std::string( buf, 4 );

    LOG_DEBUG( "Found WOWO FORM type %s", wowoform_string.c_str() );
    
    buf += 4;
    buflen -= 4;

    while ( buflen >= 8 ) {
        auto chunktype = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );

        if ( chunklen + 8 > buflen ) {
            LOG_DEBUG( "Invalid chunk length" );
            return -EINVAL;
        }

        size_t chunk_parsed_len;

        if ( chunktype == CHUNKNAME( "WTXT" ) ) {
            LOG_DEBUG( "Got WTXT chunk of %u bytes", chunklen );
            auto chrs = std::make_shared< ChunkWTXT >();
            if ( chrs->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_chunks.push_back( chrs );
        } else if ( chunktype == CHUNKNAME( "WPAR" ) ) {
            LOG_DEBUG( "Got WPAR chunk of %u bytes", chunklen );
            auto wpar = std::make_shared< ChunkWPAR >();
            if ( wpar->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_chunks.push_back( wpar );
        } else if ( chunktype == CHUNKNAME( "FORM" ) ) {
            LOG_DEBUG( "Got FORM chunk of %u bytes", chunklen );
            auto form = std::make_shared< ChunkWOWOFORM >();
            if ( form->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                return -EFAULT;
            }
            m_chunks.push_back( form );
        } else {
            auto chunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "WOWO",
                                                                                        "WTAB",
                                                                                        "LINE",
                                                                                        "RECT",
                                                                                        "WFSC" } );
            if ( chunk->parse( buf, buflen, chunk_parsed_len ) != 0 ) {
                LOG_DEBUG( "Unsupported chunk %.*s", 4, buf );
                if ( fail_on_unsupported() ) {
                    return -EINVAL;
                }
            }

            LOG_DEBUG( "got generic chunk %.*s of %u bytes", 4, buf, chunk_parsed_len );

            m_chunks.push_back( chunk );
        }

        if ( chunklen & 1 ) {
            buf += 8 + chunklen + 1;
            buflen -= 8 + chunklen + 1;
        } else {
            buf += 8 + chunklen;
            buflen -= 8 + chunklen;
        }
    }

    return 0;
}

std::string ChunkWOWOFORM::dump_as_readable_text( const shared_properties_t &shared_props )
{
    std::string res;

    for ( auto &chunk : m_chunks ) {
        res += chunk->dump_as_readable_text( shared_props );
    }

    return res;
}

json_t *ChunkWOWOFORM::dump_as_json( const shared_properties_t &shared_props,
                                     bool include_raw ) const
{
    LOG_DEBUG("called");

    auto jroot = json_object();

    json_object_set_new( jroot, "id", json_string( "FORM" ) );

    json_object_set_new( jroot, "subtype", json_string( CHUNKTYPE2NAME( m_wowoform_type ).c_str() ) );

    auto jarray = json_array();

    for ( auto &chunk : m_chunks ) {
        auto jentry = chunk->dump_as_json( shared_props, include_raw );

        json_array_append_new( jarray, jentry );
    }

    json_object_set_new( jroot, "content", jarray );

    return jroot;
}

int ChunkWOWOFORM::dump_as_binary( const shared_properties_t &shared_props,
                               std::list< IFFBinaryResult > &output,
                               bool include_raw ) const
{
    std::string parsed_text;
    std::vector< char > raw_data;
    for ( auto &chunk : m_chunks ) {
        parsed_text += chunk->dump_as_readable_text( shared_props );

        if ( include_raw ) {
            std::vector< char > tdata;
            if ( chunk->dump_as_raw( shared_props, tdata ) == 0 ) {
                for ( auto ch : tdata ) {
                    raw_data.push_back( ch );
                }
            } else {
                LOG_CRITICAL( "failed to dump raw data" );
            }
        }
    }

    IFFBinaryResult res1;
    IFFBinaryResult res2;

    res1.file_ending = "txt";
    for ( auto ch : parsed_text ) {
        res1.data.push_back( ch );
    }

    output.push_back( res1 );

    if ( include_raw ) {
        res2.file_ending = "raw";
        res2.data = raw_data;

        output.push_back( res2 );
    }

    return 0;
}
