/* find_chunk.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "find_chunk.hh"
#include "chunk_generic.hh"
#include "chunk_form.hh"
#include "chunk_cat.hh"
#include "chunk_list.hh"
#include "chunk_prop.hh"
#include "chunk_chrs.hh"
#include "chunk_anno.hh"
#include "chunk_bmhd.hh"
#include "chunk_body.hh"
#include "chunk_camg.hh"
#include "chunk_chrs.hh"
#include "chunk_cmap.hh"
#include "chunk_crng.hh"
#include "chunk_dbod.hh"
#include "chunk_dchg.hh"
#include "chunk_dest.hh"
#include "chunk_dgbl.hh"
#include "chunk_dloc.hh"
#include "chunk_dpel.hh"
#include "chunk_dpi.hh"
#include "chunk_fons.hh"
#include "chunk_grab.hh"
#include "chunk_sprt.hh"
#include "chunk_wpar.hh"
#include "chunk_wtxt.hh"
#include "chunk_name.hh"
#include "chunk_fver.hh"
#include "chunk_ssa_form.hh"
#include "chunk_cstr.hh"
#include "chunk_vhdr.hh"
#include "chunk_8svx_body.hh"
#include "chunk_lang.hh"
#include "chunk_ammd_form.hh"

#define TRY_CHUNK(name) {                                    \
    auto chunk = std::make_shared< Chunk##name >();          \
    if ( chunk->parse( buf, buflen, parsed_len ) == 0 ) {    \
        LOG_DEBUG( "Got "#name" chunk" );                    \
        return chunk;                                        \
    }                                                        \
    }

std::shared_ptr< ChunkGeneric > check_and_parse( const char *buf,
                                                 size_t buflen,
                                                 size_t &parsed_len )
{
    if ( buflen < 8 ) {
        return nullptr;
    }

    if ( buflen >= 12 ) {
        auto chunk_type = betoh_u32( get_u32( buf, 0 ) );
        auto chunklen = betoh_u32( get_u32( buf, 4 ) );
        auto form_type = betoh_u32( get_u32( buf, 8 ) );

        if ( chunk_type == CHUNKNAME( "FORM" ) &&
             form_type == CHUNKNAME( "SSA " ) ) {
            LOG_DEBUG( "special SSA FORM found" );

            if ( chunklen + 8 + 4 == buflen ) {
                LOG_DEBUG( "found SSA FORM variant 2" );
                ChunkGeneric::set_attribute( "ssa.variant", 2 );
            } else if ( chunklen + 8 + 8 + 4 == buflen ) {
                LOG_DEBUG( "found SSA FORM variant 1" );
                ChunkGeneric::set_attribute( "ssa.variant", 1 );
            }
            auto chunk = std::make_shared< ChunkSSAFORM >();
            if ( chunk->parse( buf, buflen, parsed_len ) == 0 ) {
                LOG_DEBUG( "Got SSA FORM chunk" );
                return chunk;
            }
        } else if ( chunk_type == CHUNKNAME( "FORM" ) &&
                    form_type == CHUNKNAME( "AMMD" ) ) {
            LOG_DEBUG( "special AMMD FORM found" );

            auto chunk = std::make_shared< ChunkAMMDFORM >();
            if ( chunk->parse( buf, buflen, parsed_len ) == 0 ) {
                LOG_DEBUG( "Got AMMD FORM chunk" );
                return chunk;
            }
        }
    }

    TRY_CHUNK(FORM);
    TRY_CHUNK(CAT);
    TRY_CHUNK(LIST);
    TRY_CHUNK(PROP);

    auto fillerchunk = std::make_shared< ChunkGeneric >( std::list< std::string >{ "    " } );
    if ( fillerchunk->parse( buf, buflen, parsed_len ) == 0 ) {
        LOG_DEBUG( "Got filler chunk" );
        return fillerchunk;
    }

    if ( ChunkGeneric::try_known_chunks() ) {
        TRY_CHUNK(ANNO);
        TRY_CHUNK(BMHD);
        TRY_CHUNK(BODY);
        TRY_CHUNK(CAMG);
        TRY_CHUNK(CHRS);
        TRY_CHUNK(CMAP);
        TRY_CHUNK(CRNG);
        TRY_CHUNK(DBOD);
        TRY_CHUNK(DCHG);
        TRY_CHUNK(DEST);
        TRY_CHUNK(DGBL);
        TRY_CHUNK(DLOC);
        TRY_CHUNK(DPEL);
        TRY_CHUNK(DPI);
        TRY_CHUNK(FONS);
        TRY_CHUNK(GRAB);
        TRY_CHUNK(SPRT);
        TRY_CHUNK(WPAR);
        TRY_CHUNK(WTXT);
        TRY_CHUNK(NAME);
        TRY_CHUNK(FVER);
        TRY_CHUNK(CSTR);
        TRY_CHUNK(VHDR);
        TRY_CHUNK(8SVXBODY);
        TRY_CHUNK(LANG);
    }

    if ( ! ChunkGeneric::fail_on_unsupported() ) {
        auto genchunk = std::make_shared< ChunkGeneric >();
        if ( genchunk->parse( buf, buflen, parsed_len ) == 0 ) {
            LOG_DEBUG( "Got generic chunk" );
            return genchunk;
        }
    }

    //auto chunktype = betoh_u32( get_u32( buf, 0 ) );
    LOG_DEBUG( "Unhandled chunk '%.*s'", 4, buf );
    
    return nullptr;
}
