/* main.cc
 * Copyright (C) 2020 Ralf Hoffmann.
 * ralf@boomerangsworld.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CONFIG_H_INCLUDED
#include "config.h"
#define CONFIG_H_INCLUDED
#endif // CONFIG_H_INCLUDED

#ifdef HAVE_GETOPT_LONG
#  ifndef _GNU_SOURCE
#    define _GNU_SOURCE
#    include <getopt.h>
#    undef _GNU_SOURCE
#  else
#    include <getopt.h>
#  endif
#endif
#include <cstdio>
#include <string>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <iostream>
#include <unistd.h>
#include <sstream>
#include <fstream>
#include "lowlevel_types.hh"
#include "chunk_generic.hh"
#include "ab_logging.hh"
#include "util.hh"
#include "find_chunk.hh"
#include <pwd.h>
#include <grp.h>

static void showhelp()
{
    printf( "iff-convert: IFF file converter\n");
    printf( "  Version: %s\n",PACKAGE_VERSION);
    printf( "  Author: Ralf Hoffmann <ralf@boomerangsworld.de>\n" );
    printf( "  Homepage: http://www.boomerangsworld.de/cms/tools/iff-convert\n\n" );
    printf( "  iff-convert [options] <file name>\n" );
    printf( "  Options: -V, --version\tPrint the program version and exit\n" );
    printf( "           -h, --help\t\tPrint this help and exit\n" );
    printf( "           -o, --output=<path>\tOutput file\n" );
    printf( "           -O, --output-dir=<path>\tOutput directory\n" );
    printf( "           -f, --format=<format>\tOutput format\n" );
    printf( "           \tPossible values: info, text, json, binary, extfs_copyout, extfs_list\n" );
    printf( "           -s, --set-attribute=attribute_name,value\tSet specific attributes\n" );
    printf( "           --show-attributes\tShow available attributes (needs an actual file to see attributes for the corresponding decoder)\n" );
    printf( "           --raw\tInclude raw data\n" );
    printf( "           --fail\tFail on unsupported chunks\n" );
    printf( "           --known\tTry known chunks when parsing  unsupported chunks\n" );
    printf( "           --skip-bytes=<number>>\tSkip number of bytes at the beginning of the file\n" );
    printf( "           --search-start[=<string>]>\tSearch for start of IFF file (either first four ascii characters or specific string given)\n" );
    printf( "           --extfs-copyout=<path>\tSelect file to extract for extfs\n" );
    printf( "           -d, --debug-level=<level>>\tSet debug level (0=none, 1=critical, 2=verbose, 3=debug)\n" );
    printf( "           --debug-file=<file>>\tShow debug message only from file\n" );
    printf( "           --debug-function=<function>>\tShow debug message only from function\n" );
    printf( "           <input file>\tProcess given input file\n\n" );
}

static void showversion()
{
    printf( "%s\n",PACKAGE_STRING );
    printf( "  Author: Ralf Hoffmann <ralf@boomerangsworld.de>\n" );
    printf( "  Homepage: http://www.boomerangsworld.de/cms/tools/iff-convert\n\n" );
}

class iff_parse_options {
public:
    std::string output_filename;
    std::string output_format;
    std::string output_dir;
    bool include_raw = false;
    bool show_attributes = false;
    size_t skip_bytes = 0;
    bool search_start = false;
    std::string search_string;
    std::string extfs_copyout_filename;
};

static std::string get_file_name( const iff_parse_options &options,
                                  const ChunkGeneric::IFFBinaryResult &entry,
                                  int id )
{
    char buf[128];
    snprintf( buf, sizeof( buf ), "%05d", id );

    std::string filename;
    if ( ! options.output_dir.empty() ) {
        filename = options.output_dir;
        filename += "/";
    }
    filename += "iff_dump";
    if ( ! entry.file_name_info.empty() ) {
        filename += "-";
        filename += entry.file_name_info;
        filename += "-";
    }
    filename += buf;
    if ( ! entry.file_ending.empty() ) {
        filename += ".";
        filename += entry.file_ending;
    }

    return filename;
}

static int get_user_group( std::string &username,
                           std::string &groupname )
{
    struct group *ngr;
    struct passwd *nus;

    nus = getpwuid( getuid() );
    if ( nus ) {
        username = nus->pw_name;
    }

    ngr = getgrgid( getgid() );
    if ( ngr ) {
        groupname = ngr->gr_name;
    }

    return 0;
}

static std::string get_time()
{
    char outstr[200];
    time_t t;
    struct tm *tmp;

    t = time( NULL );
    tmp = localtime( &t );
    if ( ! tmp ) {
        LOG_CRITICAL( "could not get local time" );
        return "";
    }

    if ( strftime( outstr, sizeof( outstr ), "%b %d %Y", tmp ) == 0 ) {
        LOG_CRITICAL( "failed to convert time" );
        return "";
    }

    return outstr;
}

static int output_text( const std::shared_ptr< ChunkGeneric > &chunk,
                        const iff_parse_options &options )
{
    auto res = chunk->dump_as_readable_text( {} );

    if ( options.output_filename == "-" || options.output_filename.empty() ) {
        std::cout << res << std::endl;
    } else {
        std::ofstream ofile( options.output_filename );

        ofile << res;
    }

    return 0;
}

static int output_json( const std::shared_ptr< ChunkGeneric > &chunk,
                        const iff_parse_options &options )
{
    auto json_res = chunk->dump_as_json( {}, options.include_raw );
    IFFInfo info;
    json_t *jinfo;

    int res = chunk->fill_info( info );
    if ( res == 0 ) {
        jinfo = info.get_info_as_json();
    } else {
        LOG_DEBUG( "fill info result: %d\n", res );
        jinfo = json_object();
    }

    auto json_total = json_object();
    json_object_set_new( json_total, "info", jinfo );
    json_object_set_new( json_total, "data", json_res );

    char *str1 = json_dumps( json_total, JSON_INDENT( 2 ) );

    if ( str1 ) {
        if ( options.output_filename == "-" || options.output_filename.empty() ) {
            std::cout << str1 << std::endl;
        } else {
            std::ofstream ofile( options.output_filename );

            ofile << str1;
        }
    }

    free( str1 );

    json_decref( json_total );

    return 0;
}

static int output_info( const std::shared_ptr< ChunkGeneric > &chunk,
                        const iff_parse_options &options )
{
    IFFInfo info;

    int res = chunk->fill_info( info );
    if ( res != 0 ) {
        std::cout << "Failed to get info about IFF file" << std::endl;
        return res;
    }

    auto info_str = info.get_info_as_text();

    if ( options.output_filename == "-" || options.output_filename.empty() ) {
        std::cout << info_str << std::endl;
    } else {
        std::ofstream ofile( options.output_filename );

        ofile << info_str << std::endl;
    }

    return 0;
}

static int parseBuffer( const char *buf, size_t buflen, iff_parse_options &options )
{
    size_t parsed_len;

    auto e = check_and_parse( buf, buflen, parsed_len );

    if ( ! e ) {
        LOG_CRITICAL( "Failed to parse buffer" );
        return -EINVAL;
    }

    LOG_DEBUG( "Got a valid chunk" );
    LOG_DEBUG( "Remaining bytes: %u", buflen - parsed_len );

    if ( options.output_format == "text" ) {
        output_text( e, options );
    } else if ( options.output_format == "json" ) {
        output_json( e, options );
    } else if ( options.output_format == "binary" ) {
        std::list< ChunkGeneric::IFFBinaryResult > binaries;

        int res = e->dump_as_binary( {}, binaries, options.include_raw );

        if ( res == 0 || ! ChunkGeneric::fail_on_unsupported() ) {
            if ( binaries.empty() ) {
                LOG_CRITICAL( "can't write binary data as there is none" );
            } else {
                if ( ! options.output_dir.empty() ) {
                    int id = 0;
                    for ( auto &entry : binaries ) {
                        std::string filename = get_file_name( options, entry, id++ );

                        std::ofstream ofile( filename, std::ofstream::binary );
                        if ( ofile.is_open() ) {
                            if ( entry.temp_file.empty() ) {
                                for ( auto ch : entry.data ) {
                                    ofile.put( (const char)ch );
                                }
                            } else {
                                std::ifstream infile( entry.temp_file, std::ifstream::binary );
                                char buf[ 4 * 1024 ];
                                do {
                                    infile.read( buf, sizeof( buf ) );
                                    ofile.write( buf, infile.gcount() );
                                } while ( infile.gcount() > 0 );

                                if ( ! infile.eof() ||
                                     ofile.bad() ) {
                                    LOG_CRITICAL( "failed to copy temp file" );
                                }
                            }
                        } else {
                            LOG_CRITICAL( "could not write file %s", filename.c_str() );
                        }
                    }
                } else {
                    auto main_binary = binaries.end();

                    for ( auto it = binaries.begin();
                          it != binaries.end();
                          it++ ) {
                        if ( it->main_file ) {
                            main_binary = it;
                            break;
                        }
                    }
                    
                    if ( main_binary != binaries.end() ) {
                        if ( options.output_filename == "-" || options.output_filename.empty() ) {
                            for ( auto ch : main_binary->data ) {
                                std::cout.put( (const char)ch );
                            }
                        } else {
                            std::ofstream ofile( options.output_filename, std::ofstream::binary );

                            for ( auto ch : main_binary->data ) {
                                ofile.put( (const char)ch );
                            }
                        }
                    } else {
                        LOG_CRITICAL( "can't write binary data to file/stdout because there is more than one main binary" );
                    }
                }
                if ( res != 0 ) {
                    LOG_CRITICAL( "wrote as many binary elements as possible" );
                }
            }
        } else {
            LOG_CRITICAL( "binary dump failed: %d", res );
        }
    } else if ( options.output_format == "extfs_list" ) {
        std::list< ChunkGeneric::IFFBinaryResult > binaries;
        std::string username, groupname, timestr;

        get_user_group( username, groupname );
        timestr = get_time();
        if ( timestr.empty() ) {
            timestr = "Jan 1 1970";
        }

        int res = e->dump_as_binary( {}, binaries, options.include_raw );
        if ( res == 0 ) {
            int id = 0;
            for ( auto &entry : binaries ) {
                std::string filename = "binary/";
                filename += get_file_name( options, entry, id++ );
                std::cout << "-rw-r--r-- 1 ";

                std::cout << username << " " << groupname << " ";

                if ( entry.temp_file.empty() ) {
                    std::cout << entry.data.size();
                } else {
                    std::cout << "0";
                }

                std::cout << " " << timestr << " ";
                std::cout << filename << std::endl;
            }
        }

        std::cout << "-rw-r--r-- 1 ";
        std::cout << username << " " << groupname << " ";
        std::cout << "0";
        std::cout << " " << timestr << " ";
        std::cout << "content.txt" << std::endl;

        std::cout << "-rw-r--r-- 1 ";
        std::cout << username << " " << groupname << " ";
        std::cout << "0";
        std::cout << " " << timestr << " ";
        std::cout << "info.txt" << std::endl;

        std::cout << "-rw-r--r-- 1 ";
        std::cout << username << " " << groupname << " ";
        std::cout << "0";
        std::cout << " " << timestr << " ";
        std::cout << "content.json" << std::endl;
    } else if ( options.output_format == "extfs_copyout" ) {
        if ( options.extfs_copyout_filename == "content.json" ) {
            output_json( e, options );
        } else if ( options.extfs_copyout_filename == "content.txt" ) {
            output_text( e, options );
        } else if ( options.extfs_copyout_filename == "info.txt" ) {
            output_info( e, options );
        } else {
            std::list< ChunkGeneric::IFFBinaryResult > binaries;

            int bin_res = e->dump_as_binary( {}, binaries, options.include_raw );
            if ( bin_res == 0 ) {
                int id = 0;
                bool found = false;
                for ( auto &entry : binaries ) {
                    std::string filename = "binary/";
                    filename += get_file_name( options, entry, id++ );

                    if ( filename == options.extfs_copyout_filename ) {
                        if ( options.output_filename == "-" || options.output_filename.empty() ) {
                            for ( auto ch : entry.data ) {
                                std::cout.put( (const char)ch );
                            }
                        } else {
                            std::ofstream ofile( options.output_filename, std::ofstream::binary );

                            for ( auto ch : entry.data ) {
                                ofile.put( (const char)ch );
                            }
                        }
                        found = true;
                        break;
                    }
                }

                if ( ! found ) {
                    return -ENOENT;
                }
            } else {
                return bin_res;
            }
        }
    } else {
        output_info( e, options );
    }

    return 0;
}

static int parseFile( const std::string &filename, iff_parse_options &options )
{
    std::ostringstream o;
    std::ifstream i;

    i.open( filename, std::ifstream::binary);

    if ( ! i.good() ) {
        std::cerr << "Error opening file" << std::endl;
        return -EFAULT;
    }

    o << i.rdbuf();

    std::string s1 = o.str();

    const char *data = s1.c_str();
    size_t data_len = s1.size();

    if ( options.skip_bytes > 0 ) {
        if ( data_len < options.skip_bytes ) {
            std::cerr << "File smaller than bytes to skip" << std::endl;
            return -EFAULT;
        }

        data += options.skip_bytes;
        data_len -= options.skip_bytes;
    } else if ( options.search_start ) {
        if ( ! options.search_string.empty() ) {
            const char *start = (const char*)memmem( data, data_len,
                                                     options.search_string.c_str(),
                                                     options.search_string.size() );
            if ( ! start ) {
                std::cerr << "Could not find start " << options.search_string << std::endl;
                return -EFAULT;
            }

            data_len -= start - data;
            data = start;
        } else {
            const char *start = data;
            const char *end = data + data_len;
            int alpha_found = 0;

            for ( ; start != end && alpha_found < 4; start++ ) {
                if ( isalpha( *start ) ) {
                    alpha_found++;
                } else {
                    alpha_found = 0;
                }
            }

            if ( start == end ) {
                std::cerr << "Could not find any usable start position " << std::endl;
                return -EFAULT;
            }

            start -= 4;

            data_len -= start - data;
            data = start;
        }
    }

    return parseBuffer( data, data_len, options );
}

static const int OPT_FAIL_ON_UNSUPPORTED = 256;
static const int OPT_DEBUG_FILE = 257;
static const int OPT_DEBUG_FUNCTION = 258;
static const int OPT_RAW = 259;
static const int OPT_SHOW_ATTRIBUTES = 260;
static const int OPT_TRY_KNOWN_CHUNKS = 261;
static const int OPT_SKIP_BYTES = 262;
static const int OPT_SEARCH_START = 263;
static const int OPT_EXTFS_COPYOUT = 264;

int main( int argc, char **argv )
{
    ab_set_log_level( LOG_CRITICAL );

    iff_parse_options options;

    while ( true ) {
#ifdef HAVE_GETOPT_LONG
        int option_index;
        static struct option long_options[] = {
                                               { "help", no_argument, NULL, 'h' },
                                               { "version", no_argument, NULL, 'V' },
                                               { "output", required_argument, NULL, 'o' },
                                               { "output-dir", required_argument, NULL, 'O' },
                                               { "format", required_argument, NULL, 'f' },
                                               { "fail", no_argument, NULL, OPT_FAIL_ON_UNSUPPORTED },
                                               { "raw", no_argument, NULL, OPT_RAW },
                                               { "debug-level", required_argument, NULL, 'd' },
                                               { "debug-file", required_argument, NULL, OPT_DEBUG_FILE },
                                               { "debug-function", required_argument, NULL, OPT_DEBUG_FUNCTION },
                                               { "set-attributes", required_argument, NULL, 's' },
                                               { "show-attributes", no_argument, NULL, OPT_SHOW_ATTRIBUTES },
                                               { "known", no_argument, NULL, OPT_TRY_KNOWN_CHUNKS },
                                               { "search-start", optional_argument, NULL, OPT_SEARCH_START },
                                               { "skip-bytes", required_argument, NULL, OPT_SKIP_BYTES },
                                               { "extfs-copyout", required_argument, NULL, OPT_EXTFS_COPYOUT },
                                               { NULL, 0, NULL, 0 }
        };
#endif

#ifdef HAVE_GETOPT_LONG
        int c = getopt_long( argc, argv, "hVo:d:f:O:s:", long_options, &option_index );
#else
        int c = getopt( argc, argv, "hVo:" );
#endif

        if ( c == -1 ) break;
      
        switch( c ) {
            case 'o':
                if ( optarg != NULL ) {
                    options.output_filename = optarg;
                }
                break;
            case 'O':
                if ( optarg != NULL ) {
                    options.output_dir = optarg;
                }
                break;
            case 'f':
                if ( optarg != NULL ) {
                    options.output_format = optarg;
                }
                break;
            case 'h':
                showhelp();
                exit(0);
                break;
            case 'V':
                showversion();
                exit(0);
                break;
            case OPT_FAIL_ON_UNSUPPORTED:
                ChunkGeneric::set_fail_on_unsupported( true );
                break;
            case OPT_TRY_KNOWN_CHUNKS:
                ChunkGeneric::set_try_known_chunks( true );
                break;
            case OPT_RAW:
                options.include_raw = true;
                break;
            case 'd':
                if ( optarg == NULL ) {
                    showhelp();
                    exit(1);
                } else {
                    int level = atoi( optarg );
                    switch ( level ) {
                        case 0:
                            ab_set_log_level( LOG_NOTHING );
                            break;
                        case 1:
                            ab_set_log_level( LOG_CRITICAL );
                            break;
                        case 2:
                            ab_set_log_level( LOG_VERBOSE );
                            break;
                        case 3:
                            ab_set_log_level( LOG_DEBUG );
                            break;
                        default:
                            showhelp();
                            exit(1);
                    }
                }
                break;
            case OPT_DEBUG_FILE:
                if ( ! optarg ) {
                    showhelp();
                    exit(1);
                } else {
                    ab_add_file_filter( optarg );
                }
                break;
            case OPT_DEBUG_FUNCTION:
                if ( ! optarg ) {
                    showhelp();
                    exit(1);
                } else {
                    ab_add_function_filter( optarg );
                }
                break;
            case 's':
                if ( optarg != NULL ) {
                    std::vector< std::string > attr;
                    split_string( attr, optarg, '=' );
                    if ( attr.size() != 2 ) {
                        showhelp();
                        exit( 1 );
                    }

                    ChunkGeneric::set_attribute( attr[0], atoi( attr[1].c_str() ) );
                }
                break;
            case OPT_SHOW_ATTRIBUTES:
                options.show_attributes = true;
                break;
            case OPT_SEARCH_START:
                options.search_start = true;
                if ( optarg != NULL ) {
                    options.search_string = optarg;
                }
                break;
            case OPT_SKIP_BYTES:
                if ( optarg == NULL ) {
                    showhelp();
                    exit(1);
                } else {
                    int s = atoi( optarg );
                    if ( s < 0 ) {
                        showhelp();
                        exit(1);
                    }

                    options.skip_bytes = s;
                }
                break;
            case OPT_EXTFS_COPYOUT:
                if ( optarg == NULL ) {
                    showhelp();
                    exit(1);
                } else {
                    options.extfs_copyout_filename = optarg;
                    options.output_format = "extfs_copyout";
                }
                break;
            default:
                break;
        }
    }

    if ( optind >= argc ) {
        showhelp();
        exit(0);
    }

    int res = parseFile( argv[optind], options );
    LOG_DEBUG( "Result: %d", res );

    if ( options.show_attributes ) {
        auto attrs = ChunkGeneric::get_attributes();

        std::cout << "Attributes:" << std::endl;
        for ( auto &a : attrs ) {
            std::cout << a.first << "=" << a.second << std::endl;
        }
    }

    if ( res != 0 ) {
        exit( EXIT_FAILURE );
    }
            
    return 0;
}
